# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import argparse
import os
import sys
import traceback
from logging import error, info
from typing import Optional, cast

from portmod.config import config_to_string, get_config
from portmod.config.sets import get_system
from portmod.globals import env, get_version
from portmod.loader import load_installed_pkg
from portmod.modules import handle_cfg_protect
from portmod.prefix import get_prefixes
from portmod.repos import get_local_repos
from portmod.sync import sync
from portmod.transactions import get_usestrings
from portmodlib.atom import Atom
from portmodlib.l10n import l10n
from portmodlib.log import add_logging_arguments, init_logger

from .conflicts import add_conflicts_parser
from .error import InputException
from .init import add_init_parser
from .merge import add_merge_parser
from .mirror import add_mirror_parser
from .query import add_query_parser
from .search import add_search_parser
from .select import add_select_parser
from .use import add_use_parser
from .validate import validate


def add_info_parser(subparsers, parents):
    def info(args):
        from git import Repo

        # Strip the user's home directory from paths and replace it with ~
        # This command is designed to produce information to be submitted with
        # bug reports, and the user's home directory is not a relevant piece of information.
        def strip_user(string: str) -> str:
            return string.replace(os.path.expanduser("~"), "~")

        print(f"Portmod {get_version()}")
        print(f"Python {sys.version}")
        print()

        print(l10n("info-repositories"))
        for repo in env.prefix().REPOS:
            gitrepo = Repo.init(repo.location)
            print("    ", "\n         ".join(strip_user(str(repo)).split(",")))
            print(
                "    ",
                l10n(
                    "info-repository-date", date=gitrepo.head.commit.committed_datetime
                ),
            )
            print("    ", l10n("info-repository-commit", commit=gitrepo.head.commit))
            print()

        generator = (
            load_installed_pkg(atom)
            for atom in get_system() | set(map(Atom, get_config()["INFO_PACKAGES"]))
            if load_installed_pkg(atom)
        )
        packages = set(filter(None, generator))
        length = 0
        if len(packages) > 0:
            length = max(len(x.CPN) for x in packages)
        for pkg in sorted(packages, key=lambda x: x.CPN):
            # FIXME: Display date package was installed for live packages
            usestrings = filter(None, get_usestrings(pkg, pkg.INSTALLED_USE, True))
            padding = length - len(str(pkg.CPN))
            print(pkg.CPN + ":", " " * padding, pkg.PVR, *usestrings)
        print()

        # Print config values
        config = get_config()
        if args.verbose:
            config_string = config_to_string(config)
        else:
            config_string = config_to_string(
                {
                    entry: config[entry]
                    for entry in config
                    if entry in config["INFO_VARS"]
                }
            )

        print(strip_user(config_string))
        # Print hardcoded portmod paths
        print(strip_user(f"TMP_DIR = {env.TMP_DIR}"))
        print(strip_user(f"CACHE_DIR = {env.CACHE_DIR}"))
        print(strip_user(f"CONFIG_DIR = {env.prefix().CONFIG_DIR}"))
        print(strip_user(f"ROOT = {env.prefix().ROOT}"))
        sys.exit(0)

    parser = subparsers.add_parser("info", help=l10n("info-help"), parents=parents)
    parser.set_defaults(func=info)


def parse_args(
    subcommand: Optional[str] = None, prefix_subcommand: Optional[str] = None
):
    common = argparse.ArgumentParser(add_help=False)
    add_logging_arguments(common)
    common.add_argument("--debug", help=l10n("merge-debug-help"), action="store_true")

    parser = argparse.ArgumentParser(description=l10n("description"), parents=[common])
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--version", help=l10n("version-help"), action="store_true")

    subparsers = parser.add_subparsers(dest="subcommand_name")

    if subcommand is None or subcommand in get_prefixes() and prefix_subcommand is None:
        prefix_parsers_list = []

        def get_help_func(parser):
            func = parser.print_help
            return lambda args: func()

        for prefix in get_prefixes():
            prefix_parser = subparsers.add_parser(
                prefix, help=l10n("prefix-help", prefix=prefix), parents=[common]
            )
            prefix_parsers_list.append(
                prefix_parser.add_subparsers(dest="prefix_subcommand_name")
            )
            prefix_parser.set_defaults(func=get_help_func(prefix_parser))

        for subparser in prefix_parsers_list:
            subparser.add_parser("merge", help=l10n("merge-help"), add_help=False)
            subparser.add_parser("search", help=l10n("search-help"), add_help=False)
            subparser.add_parser("select", help=l10n("select-help"), add_help=False)
            subparser.add_parser("query", help=l10n("query-help"), add_help=False)
            subparser.add_parser("use", help=l10n("use-help"), add_help=False)
            subparser.add_parser(
                "conflict-ui", help=l10n("conflict-ui-help"), add_help=False
            )
            subparser.add_parser("info", help=l10n("info-help"), add_help=False)
            subparser.add_parser("validate", help=l10n("validate-help"), add_help=False)
            subparser.add_parser(
                "migrate",
                help="Migration tool for new flat filesystem layout",
                add_help=False,
            )
        subparsers.add_parser("mirror", help=l10n("mirror-help"), add_help=False)
    elif subcommand in get_prefixes():
        prefix = subcommand

        prefix_parser = subparsers.add_parser(
            prefix, help=l10n("prefix-help", prefix=prefix), parents=[common]
        )
        prefix_parsers = prefix_parser.add_subparsers(dest="prefix_subcommand_name")
        prefix_parser.set_defaults(func=lambda args: prefix_parser.print_help())
        if prefix_subcommand == "merge":
            add_merge_parser(prefix_parsers, [common])
        elif prefix_subcommand == "search":
            add_search_parser(prefix_parsers, [common])
        elif prefix_subcommand == "select":
            add_select_parser(prefix_parsers, [common])
        elif prefix_subcommand == "query":
            add_query_parser(prefix_parsers, [common])
        elif prefix_subcommand == "use":
            add_use_parser(prefix_parsers, [common])
        elif prefix_subcommand == "conflict-ui":
            add_conflicts_parser(prefix_parsers, [common])
        elif prefix_subcommand == "info":
            add_info_parser(prefix_parsers, [common])
        elif prefix_subcommand == "validate":
            prefix_parsers.add_parser(
                "validate", help=l10n("validate-help"), parents=[common]
            ).set_defaults(func=validate)
        elif prefix_subcommand == "migrate":
            prefix_parsers.add_parser(
                "migrate", help="Migration tool for new flat filesystem layout"
            ).set_defaults(func=flat_migrate)
    elif subcommand == "mirror":
        add_mirror_parser(subparsers, [common])

    add_init_parser(subparsers, parents=[common])
    subparsers.add_parser(
        "sync", help=l10n("sync-help"), parents=[common]
    ).set_defaults(func=lambda args: sync(get_local_repos().values()))

    try:
        import argcomplete  # pylint: disable=import-error

        argcomplete.autocomplete(parser)
    except ModuleNotFoundError:
        pass

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(2)

    if "--ignore-default-opts" in sys.argv:
        args = sys.argv[1:]
    else:
        args = sys.argv[1:] + os.environ.get("OMWMERGE_DEFAULT_OPTS", "").split()

    if subcommand is None:
        return parser.parse_known_args(args)[0]
    else:
        return parser.parse_args(args)


# FIXME: Remove
# Migration code for non-flat filesystems
def flat_migrate(_args):
    import shutil

    from portmod.loader import load_all_installed
    from portmod.parsers.manifest import FileType, Manifest, ManifestEntry

    print(
        "Installed files will be restructured, and files installed by each package"
        " will be registered."
    )

    if not get_config().get("INSTALL_DEST"):
        error(
            "You need to select a new profile, otherwise the "
            "resulting installation may not be functional"
        )
        sys.exit(1)

    path = os.path.join(env.prefix().ROOT, "modules")
    # Modules set is no longer used
    if os.path.exists(path):
        os.remove(path)
    os.makedirs(env.prefix().VARIABLE_DATA, exist_ok=True)
    os.makedirs(env.prefix().VAR_SET_DIR, exist_ok=True)

    for entry in ["rebuild", "world", "world_sets"]:
        if os.path.exists(os.path.join(env.prefix().ROOT, entry)):
            shutil.move(
                os.path.join(env.prefix().ROOT, entry),
                os.path.join(env.prefix().VAR_SET_DIR, entry),
            )

    for entry in ["vfs", "vfs-archives"]:
        if os.path.exists(os.path.join(env.prefix().ROOT, "pkg", entry)):
            shutil.move(
                os.path.join(env.prefix().ROOT, "pkg", entry),
                os.path.join(env.prefix().VARIABLE_DATA, entry),
            )

    for entry in ["db", "news", "module-data"]:
        if os.path.exists(os.path.join(env.prefix().ROOT, entry)):
            shutil.move(
                os.path.join(env.prefix().ROOT, entry),
                os.path.join(env.prefix().VARIABLE_DATA, entry),
            )

    for pkg in load_all_installed():
        if "local" not in pkg.PROPERTIES:
            contents_manifest = Manifest()
            info(f"Registering files for package {pkg}")

            pkg_path = os.path.join(env.prefix().ROOT, "pkg", pkg.CATEGORY, pkg.PN)
            if os.path.islink(pkg_path):
                contents_manifest.add_entry(
                    ManifestEntry.from_path(
                        FileType.MISC,
                        pkg_path,
                        os.path.join("pkg", pkg.CATEGORY, pkg.PN),
                    )
                )
            else:
                for (path, _, filenames) in os.walk(pkg_path, followlinks=False):
                    for name in filenames:
                        contents_manifest.add_entry(
                            ManifestEntry.from_path(
                                FileType.MISC,
                                os.path.join(path, name),
                                os.path.relpath(
                                    os.path.join(path, name), env.prefix().ROOT
                                ),
                            )
                        )

            db_path = os.path.join(env.prefix().INSTALLED_DB, pkg.CATEGORY, pkg.PN)
            contents_manifest.write(os.path.join(db_path, "CONTENTS"))

    sys.exit(0)


def is_old_structure() -> bool:
    """
    Determine if installation data from the alpha/beta releases
    are installed on this system
    """
    return os.path.exists(os.path.join(env.DATA_DIR, "mods"))


def migrate():
    """
    Migrates installation data from alpha/beta to new system
    """
    # TODO: Remove this migration code at before the 2.0 release
    # This is NOT a feature, but rather a temporary convenience
    # Post 2.0, users who still want to migrate installations from alpha/beta should
    # first install a release candidate to make use of the migration code,
    # then switch to the latest release
    import re
    import shutil
    from logging import warning

    from portmod.config import set_config_value
    from portmod.prefix import add_prefix
    from portmod.prompt import prompt_bool, prompt_str
    from portmod.repos import get_local_repos
    from portmod.sync import sync

    if not prompt_bool(
        """
    Old data from the alpha or beta version of portmod has been detected.
    Would you like to automatically migrate this data to the new system?
    There will not be an option later to reverse this automatically (though it is possible to do it manually)
    """
    ):
        sys.exit(1)

    prefix = prompt_str(
        """What would you like to call your new prefix?

This will be used in commands (e.g. portmod <prefix> merge) to identify the prefix.

It is also recommended that you create aliases for your prefixes, such as (in bash):
alias omwmerge="portmod openmw merge"
    """,
        "openmw",
    )

    if prefix is None:
        sys.exit(1)

    if os.path.exists(os.path.join(env.DATA_DIR, "openmw")):
        # If they use the (old) default location for the openmw repo
        # Move it into the new default location
        os.makedirs(env.REPOS_DIR, exist_ok=True)
        shutil.move(
            os.path.join(env.DATA_DIR, "openmw"), os.path.join(env.REPOS_DIR, "openmw")
        )
        # If path is in repos.cfg, rewrite it
        import configparser

        repo_config = configparser.ConfigParser(
            comment_prefixes="/", allow_no_value=True
        )
        repo_config.read(env.REPOS_FILE)
        if "openmw" in repo_config:
            repo_config.set("openmw", "location", os.path.join(env.REPOS_DIR, "openmw"))
            with open(env.REPOS_FILE, "w") as repos_file:
                repo_config.write(repos_file)

    meta_repo = get_local_repos()["meta"]
    sync([meta_repo])

    # Create and enter openmw prefix
    add_prefix(prefix, get_config().get("ARCH") or "openmw")
    env.set_prefix(prefix)

    shutil.move(
        os.path.join(env.DATA_DIR, "mods"), os.path.join(env.prefix().ROOT, "pkg")
    )

    # Move old data into new prefix
    data_files = set(os.listdir(env.DATA_DIR)) - {"repos", prefix, "prefix"}
    if os.path.exists(env.prefix().INSTALLED_DB):
        # db has already been initialized as part of prefix initialization
        shutil.rmtree(env.prefix().INSTALLED_DB)
    os.makedirs(env.prefix().ROOT, exist_ok=True)
    for file in data_files:
        shutil.move(
            os.path.join(env.DATA_DIR, file),
            os.path.join(env.prefix().ROOT, file),
        )

    # Move old config files into new prefix
    config_files = set(os.listdir(env.CONFIG_DIR)) - {"repos.cfg", prefix}
    os.makedirs(env.prefix().CONFIG_DIR, exist_ok=True)
    for file in config_files:
        if file.startswith("mod."):
            shutil.move(
                os.path.join(env.CONFIG_DIR, file),
                os.path.join(
                    env.prefix().CONFIG_DIR, re.sub(r"^mod\.", "package.", file)
                ),
            )
        else:
            shutil.move(
                os.path.join(env.CONFIG_DIR, file),
                os.path.join(env.prefix().CONFIG_DIR, file),
            )

    # Delete old pybuild cache
    if os.path.exists(os.path.join(env.CACHE_DIR, "pybuild")):
        shutil.rmtree(os.path.join(env.CACHE_DIR, "pybuild"))

    # Move old cfg_protect dir into prefix
    if os.path.exists(os.path.join(env.CACHE_DIR, "cfg_protect")):
        os.makedirs(env.CACHE_DIR, exist_ok=True)
        shutil.move(
            os.path.join(env.CACHE_DIR, "cfg_protect"), env.prefix().CONFIG_PROTECT_DIR
        )

    set_config_value("REPOS", "openmw")

    warning(
        "Exiting portmod so that settings can be reloaded properly. "
        "While this could be done in-process, it's more work than it's worth\n"
        "Your previous command has not been executed. "
        "If it was important, you will need to run it again."
    )
    sys.exit(1)


def main():
    os.environ["PYTHONUNBUFFERED"] = "1"

    if is_old_structure():
        migrate()

    args = parse_args()
    init_logger(args)

    if args.subcommand_name in get_prefixes():
        env.set_prefix(args.subcommand_name)

        # Ensure that we read config entries into os.environ
        get_config()

        if os.path.exists(os.path.join(env.prefix().SET_DIR, "world")):
            from portmod.atom import Atom
            from portmod.config.sets import add_set
            from portmod.parsers.list import read_list

            # FIXME: Remove
            # Bug in 2.0_rc0 and 2.0_rc1 caused selected packages to be written to
            # the wrong file
            for atom in read_list(os.path.join(env.prefix().SET_DIR, "world")):
                add_set("selected-packages", Atom(atom))
            os.remove(os.path.join(env.prefix().SET_DIR, "world"))

    prefix_subcommand_name = None
    if hasattr(args, "prefix_subcommand_name"):
        prefix_subcommand_name = cast(str, args.prefix_subcommand_name)
    args = parse_args(args.subcommand_name, prefix_subcommand_name)

    env.DEBUG = args.debug

    init_logger(args)

    if args.version:
        print(f"Portmod {get_version()}")
        sys.exit(0)

    try:
        args.func(args)
        if args.subcommand_name in get_prefixes():
            handle_cfg_protect()
    except InputException as e:
        error("{}".format(e))
        sys.exit(1)
    except Exception as e:
        traceback.print_exc()
        error("{}".format(e))
        sys.exit(1)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Optional


def vdb_path() -> str:
    return os.environ["PORTMOD_INSTALLED_DB"]


def prefix_name() -> Optional[str]:
    return os.environ.get("PORTMOD_PREFIX_NAME")


def download_dir() -> str:
    return os.environ["PORTMOD_DOWNLOAD_DIR"]


def messages_dir() -> str:
    return os.environ["PORTMOD_MESSAGES_DIR"]


def warnings_dir() -> str:
    return os.environ["PORTMOD_WARNINGS_DIR"]


def root() -> str:
    return os.environ["PORTMOD_ROOT"]


def config_protect_dir() -> str:
    return os.environ["PORTMOD_CONFIG_PROTECT_DIR"]



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
The module accessible within pybuilds

Note that this module should not be imported outside of pybuild files
"""

from portmodlib.atom import version_gt  # noqa  # pylint: disable=unused-import
from portmodlib.fs import patch_dir  # noqa  # pylint: disable=unused-import
from portmodlib.globals import download_dir as _download_dir
from portmodlib.masters import get_masters  # noqa  # pylint: disable=unused-import
from portmodlib.pybuild import File, InstallDir  # noqa  # pylint: disable=unused-import
from portmodlib.usestr import use_reduce  # noqa  # pylint: disable=unused-import
from portmodlib.vfs import find_file, list_dir  # noqa  # pylint: disable=unused-import

from .pybuild import Pybuild1, apply_patch  # noqa  # pylint: disable=unused-import

DOWNLOAD_DIR = _download_dir()

__all__ = [
    "version_gt",
    "patch_dir",
    "get_masters",
    "use_reduce",
    "find_file",
    "list_dir",
    "File",
    "InstallDir",
    "Pybuild1",
    "apply_patch",
    "DOWNLOAD_DIR",
]



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Dependency resolution module

Converts mod dependency links and REQUIRED_USE conditions into a MAX-SAT formula in
conjunctive normal form.
This formula is then solved using pysat (python-sat on pypi) and the resulting model
converted back into a list of installed mods and their use flag configuration.

Note that the hard requirements defined in DEPEND, RDEPEND and REQUIRED_USE are
converted into a SAT formula that must be solved in its entirety.
We use a MAX-SAT solver because there are also other soft requirements which are used
to avoid installing mods unnecessarily and to avoid changing the user's use flag
configuration, if possible.

See https://en.wikipedia.org/wiki/Boolean_satisfiability_problem for details on
the SAT problem
"""

import re
from logging import info
from typing import AbstractSet, Dict, Iterable, List, Set

from portmod.config.sets import get_set
from portmod.config.use import get_use
from portmod.loader import AmbiguousAtom, load_all_installed, load_pkg
from portmod.transactions import (
    PackageDoesNotExist,
    Transactions,
    UseDep,
    generate_transactions,
)
from portmodlib.atom import Atom, FQAtom, QualifiedAtom, atom_sat
from portmodlib.l10n import l10n

from .formula import Formula, Target, add_masked, generate_formula
from .weights import weigh_clauses


class DepError(Exception):
    """Indicates an unsatisfiable transaction"""


def resolve(
    enabled: Iterable[Atom],
    disabled: Iterable[Atom],
    explicit: AbstractSet[Atom],
    selected: AbstractSet[Atom],
    selected_sets: AbstractSet[str],
    *,
    deep: bool = False,
    noreplace: bool = False,
    update: bool = False,
    newuse: bool = False,
    depclean: bool = False,
    emptytree: bool = False,
) -> Transactions:
    """
    Calculates new mod configuration to match system after the given mods are installed

    Note: We have two modes of operation:

    Shallow
        We assume that all installed mods are fixed and will not
        change version. Any version of a newly selected mods may be installed.
        Note that use flags may change on installed mods.
    Deep
        We treat every mod as newly selected, and choose from among its versions

    args:
        enabled: Packages which are to be enabled/installed
        disabled: Packages which are to be disabled/removed
        selected: Enabled packages which were explicitly selected
        selected_sets: Sets which were explicitly selected (the contents of which
                would be in enabled)
        deep: Whether or not we are running in deep mode
        noreplace: Whether or not explicitly selected packages and sets should be rebuilt,
                   even if no other changes to them are necessary
        update: If true, packages will be updated, if possible.
        newuse: If true, packages with changes to their use flags will be rebuilt.
        depclean: If true, packages which were neither explicitly selected, nor
                  required as dependencies, will be removed.
        emptytree: If true, all packages in the dependency tree will be rebuilt, as
                   if nothing was installed.
    returns:
        Transactions object representing the package changes required
    """
    # Slow imports
    from pysat.examples.rc2 import RC2
    from pysat.solvers import Solver

    info(l10n("calculating-dependencies"))
    formula = Formula()

    # List of sets of mod objects, with each being a specific version of that mod
    oldselected: List[Target] = []
    newenabled: Dict[str, Target] = dict()

    CMD_ATOM = "mods passed on command line"
    WORLD_ATOM = "world favourites file"

    for atom in list(enabled) + list(disabled):
        if not load_pkg(atom):
            raise PackageDoesNotExist(atom)

    newenabledset = {
        atom: CMD_ATOM
        for atom in set(enabled)
        | {atom for set_name in selected_sets for atom in get_set(set_name)}
    }
    for atom in disabled:
        name = load_pkg(atom)[0].CPN
        if name in newenabledset:
            del newenabledset[name]

    for atom in disabled:
        for mod in load_pkg(atom):
            formula.append_dep(["-" + mod.ATOM], CMD_ATOM, atom)

    def create_modlist(atom):
        modlist = load_pkg(atom)

        # Raise exception if mod name is ambiguous (exists in multiple categories)
        if not all(mod.ATOM.C == modlist[0].ATOM.C for mod in modlist):
            raise AmbiguousAtom(atom, [mod.ATOM for mod in modlist])

        if not modlist:
            if atom in set(selected):
                raise PackageDoesNotExist(atom)

            raise PackageDoesNotExist(
                msg=l10n("package-does-not-exist-in-world", atom=atom)
            )
        return modlist

    for atom, source in newenabledset.items():
        modlist = create_modlist(atom)
        name = modlist[0].CPN
        if name in newenabled:
            newenabled[modlist[0].CPN].pkgs.extend(modlist)
            if newenabled[modlist[0].CPN].source is CMD_ATOM or source is CMD_ATOM:
                # Use generic atom if included multiple times on command line.
                # Not all versions in modlist will correspond to a specific version
                # passed on the command line.
                newenabled[modlist[0].CPN].atom = name
            # Prefer command line as source rather than world file
            if newenabled[modlist[0].CPN].source is WORLD_ATOM and source is CMD_ATOM:
                newenabled[modlist[0].CPN].source = CMD_ATOM
        else:
            newenabled[modlist[0].CPN] = Target(modlist, atom, source)

    for atom in get_set("world") - {load_pkg(atom)[0].CPN for atom in disabled}:
        if atom not in selected:
            modlist = create_modlist(atom)
            oldselected.append(Target(modlist, atom, "world favourites file"))

    # Any remaining installed mods don't need to remain installed if there aren't
    # any dependencies, so source is None
    installed = [Target([mod], mod.ATOM, None) for mod in load_all_installed()]

    selected_cpn = set()
    explicit_cpn = set()

    for atom in explicit:
        pkg = load_pkg(atom)[0]
        explicit_cpn.add(pkg.CPN)

    for atom in selected:
        pkg = load_pkg(atom)[0]
        selected_cpn.add(pkg.CPN)

    selected_from_sets: Set[QualifiedAtom] = set()
    for set_name in selected_sets:
        selected_from_sets |= {QualifiedAtom(atom) for atom in get_set(set_name)}

    depsadded: Set[FQAtom] = set()
    # Hard clauses
    formula.merge(
        generate_formula(list(newenabled.values()) + oldselected + installed, depsadded)
    )
    # Soft clauses
    formula.merge(
        weigh_clauses(
            formula.atoms,
            formula.flags,
            explicit=explicit_cpn,
            deep=deep,
            depclean=depclean,
            update=update,
            newuse=newuse,
            emptytree=emptytree,
        )
    )

    formula.merge(add_masked(formula.atoms, formula.flags))

    if depclean:
        for pkg in load_all_installed():
            # When depcleaning, installed packages should be frozen at their current
            # version, or else removed
            versions = load_pkg(Atom(pkg.ATOM.CPN))
            var = formula.genvariable(
                [f"No non-installed versions of {pkg.CPN} are allowed"]
            )
            # Clause requires that either the installed version should be removed,
            # or all other versions should be removed.
            # This means it is not possible for a version other than the installed version
            # to be kept
            formula.append_dep(
                ["-" + pkg.ATOM, var],
                "Selected packages must not be changed when depcleaning",
                pkg.ATOM,
            )
            # Var can only be true if all versions are not installed
            # as each clause requires that either a particular version is not installed,
            # or var is false.
            # Hence if var is true, every one of these clauses is satisfied by the left term
            for other in versions:
                if not other.INSTALLED:
                    formula.append(
                        ["-" + other.ATOM, "-" + var],
                        other.ATOM,
                        "Selected packages must not be changed when depcleaning",
                    )

            for flag in pkg.IUSE_EFFECTIVE:
                if flag in get_use(pkg)[0]:
                    formula.append_usedep(
                        [pkg.ATOM + "[" + flag + "]", "-" + pkg.ATOM],
                        "Selected packages must not be changed when depcleaning",
                        pkg.ATOM,
                        flag,
                    )
                else:
                    formula.append_usedep(
                        ["-" + pkg.ATOM + "[" + flag + "]", "-" + pkg.ATOM],
                        "Selected packages must not be changed when depcleaning",
                        pkg.ATOM,
                        "-" + flag,
                    )

    if not formula.clauses:
        return Transactions()

    wcnf = formula.get_wcnfplus()
    solver = RC2(wcnf, solver="mc")
    solver.compute()
    if solver.compute():
        info(l10n("done"))
        # Turn numbers in result back into strings
        result = list(
            filter(
                # Filter out custom variables that are only meaningful
                # for the computation
                lambda x: not x.startswith("_") and not x.startswith("-_"),
                [Formula.getstring(num) for num in solver.model],
            )
        )
        flags = [atom for atom in result if "[" in atom]
        enabled_final = [
            FQAtom(mod) for mod in result if "[" not in mod and mod[0] != "-"
        ]
        enablednames = [atom.CPN for atom in enabled_final]
        disabled_final = [
            FQAtom(mod.lstrip("-"))
            for mod in result
            if "[" not in mod and mod[0] == "-"
            # If mod is enabled and installed version is disabled,
            # ignore disabled version, and vice versa
            and Atom(mod.lstrip("-")).CPN not in enablednames
        ]
        usedeps = []

        for flag in flags:
            atom = FQAtom(flag.lstrip("-"))
            if flag[0] == "-":
                prefix = "-"
            else:
                prefix = ""

            # Ignore flags for mods that are to be uninstalled
            if re.sub(r"\[.*\]", "", atom) in enabled_final:
                usedeps.append(
                    UseDep(atom.strip_use(), prefix + list(atom.USE)[0], None)
                )

        transactions = generate_transactions(
            enabled_final,
            disabled_final,
            selected_cpn,
            usedeps,
            noreplace=noreplace,
            emptytree=emptytree,
            update=update,
        )
        return transactions

    # Find clause that caused the solver to fail
    # Backtrack mod that added that clause until we reach @selected by
    # looking for clauses containing the (enabled) mod as a token.
    # Display this trace.
    # Then also display trace for a mod that requires/blocks the clause that
    # caused the failure

    solver2 = Solver("mc")
    solveableformula = []
    # Add atmost clauses first, as they won't by themselves cause conflicts,
    # and are not very useful for explaining a failed transaction
    for clause in formula.get_clauses():
        if isinstance(clause, Formula.MetaClause) and clause.atmost is not None:
            solver2.add_atmost(clause.intclause, clause.atmost)

    model = set()
    for clause in formula.get_clauses():
        if isinstance(clause, Formula.MetaClause) and (
            clause.atmost is not None or clause.weight is not None
        ):
            continue
        else:
            solver2.add_clause(clause.intclause)

        if solver2.solve():
            model = set(map(Formula.getstring, solver2.get_model()))
            solveableformula.append(clause)
        else:
            conflict = None

            for solveableclause in solveableformula:
                # Find clause that contradicts failed clause
                # Note that metaclauses don't have a blocks function,
                if hasattr(solveableclause, "blocks") and solveableclause.blocks(
                    model, clause
                ):
                    conflict = solveableclause
                    break

            def find_dependency(modstr: str):
                # Ignore top-level dependencies
                if not isinstance(modstr, Atom):
                    return None

                for solveableclause in solveableformula:
                    if isinstance(solveableclause, Formula.DepClause):
                        if atom_sat(Atom(modstr), solveableclause.dependency) and all(
                            var in model for var in solveableclause.requirements
                        ):
                            return solveableclause
                return None

            exceptionstring = f"{clause}\n"
            parent = find_dependency(clause.source)
            i = 1
            while parent is not None:
                exceptionstring += i * "  " + f"{parent}\n"
                parent = find_dependency(parent.source)
                i += 1

            if conflict:
                exceptionstring += l10n("contradicts") + "\n"
                exceptionstring += f"{conflict}\n"
                parent = find_dependency(conflict.source)
                i = 1
                while parent is not None:
                    exceptionstring += i * "  " + f"{parent}\n"
                    parent = find_dependency(parent.source)
                    i += 1

            raise DepError(
                l10n("unable-to-satisfy-dependencies") + f"\n{exceptionstring}"
            )

    raise Exception(
        "Internal error: Unable to satisfy dependencies, "
        "but there are no clauses in the formula!"
    )



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
import stat
from functools import lru_cache
from shutil import copystat
from typing import Callable, Generator, List, Optional, Set, Tuple, Union

from portmodlib.portmod import _get_hash
from portmodlib.source import HashAlg

# 32MB buffer seems to give the best balance between performance on large files
# and on small files
HASH_BUF_SIZE = 32 * 1024 * 1024


def onerror(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise  # pylint: disable=misplaced-bare-raise


def _move2(src: os.DirEntry, dest: str):
    if os.path.islink(src):
        return os.symlink(os.readlink(src), dest)
    return shutil.move(src.path, dest)


def _patch_file(
    src: os.DirEntry,
    dst: str,
    overwrite: bool = True,
    move_function: Callable[[os.DirEntry, str], None] = _move2,
):
    if os.path.exists(dst) and src.is_file():
        if overwrite:
            os.remove(dst)
        else:
            raise FileExistsError(f"File {dst} already exists")

    move_function(src, dst)


def _iter_files_to_patch(
    src: Union[str, os.DirEntry],
    dst: str,
    *,
    ignore: Callable[[str, List[str]], Set[str]] = None,
    case_sensitive: bool = True,
) -> Generator[Tuple[os.DirEntry, str], None, None]:
    with os.scandir(src) as itr:
        entries = list(itr)
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    for entry in entries:
        if entry.name in ignored_names:
            continue
        if case_sensitive:
            dstname = os.path.join(dst, entry.name)
        else:
            dstname = ci_exists(os.path.join(dst, entry.name)) or os.path.join(
                dst, entry.name
            )

        if entry.is_symlink():
            yield (entry, dstname)
        elif entry.is_dir():
            yield from _iter_files_to_patch(
                entry,
                dstname,
                ignore=ignore,
                case_sensitive=case_sensitive,
            )
        else:
            yield (entry, dstname)


# Modified version of shutil.copytree from
# https://github.com/python/cpython
# Python software and documentation are licensed under the
# Python Software Foundation License Version 2
def patch_dir(
    src: Union[str, os.DirEntry],
    dst: str,
    *,
    overwrite: bool = True,
    ignore: Callable[[str, List[str]], Set[str]] = None,
    case_sensitive: bool = True,
    move_function: Callable[[os.DirEntry, str], None] = _move2,
) -> str:
    """
    Copies src ontop of dst

    args:
        src: Source directory to copy from
        dst: Destination directory to copy to
        overwrite: If true, overwrite existing files.
        ignore: A callable which, given a directory and its contents, should return
            a set of files to ignore
        case_sensitive: If False, treat file and directory names as case insensitive
        move_function: The function to use to transfer individual files.
            Default is shutil.move (modified to accept a DirEntry).
            The signature should match shutil.copy2.

    raises:
        FileExistsError

    returns:
        Returns dst
    """

    for src_file, dst_file in _iter_files_to_patch(
        src,
        dst,
        ignore=ignore,
        case_sensitive=case_sensitive,
    ):
        parent_dir = os.path.dirname(dst_file)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)
            try:
                copystat(os.path.dirname(src_file.path), parent_dir)
            except OSError as why:
                if getattr(why, "winerror", None) is None:
                    raise why

        _patch_file(
            src_file,
            dst_file,
            overwrite=overwrite,
            move_function=move_function,
        )

    return dst


def ci_exists(path: str) -> Optional[str]:
    """
    Checks if a path exists, ignoring case.

    If the path exists but is ambiguous the result is not guaranteed
    """
    partial_path = "/"
    for component in os.path.normpath(os.path.abspath(path)).split(os.sep)[1:]:
        found = False
        for entryname in os.listdir(partial_path):
            if entryname.lower() == component.lower():
                partial_path = os.path.join(partial_path, entryname)
                found = True
                break
        if not found:
            return None

    if os.path.exists(partial_path):
        return partial_path

    return None


def get_tree_size(path):
    """Return total size of files in given path and subdirs."""
    total = 0
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            total += get_tree_size(entry.path)
        else:
            total += entry.stat(follow_symlinks=False).st_size
    return total


@lru_cache(maxsize=None)
def get_hash(filename: str, funcs=(HashAlg.BLAKE3,)) -> List[str]:
    """Hashes the given file"""
    return _get_hash(filename, [func.value for func in funcs], HASH_BUF_SIZE)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import pickle
import tempfile
from functools import lru_cache
from typing import Any, List, Optional

from appdirs import AppDirs

from portmodlib.l10n import l10n


@lru_cache()
def get_version() -> str:
    """Returns portmod version"""
    try:
        from ._version import version

        return str(version)
    except ImportError:
        osp = os.path
        possible_root = osp.dirname(osp.dirname(osp.realpath(__file__)))
        if osp.isfile(osp.join(possible_root, ".portmod_not_installed")):
            # Only a dev dependency and is otherwise unneeded at runtime
            import setuptools_scm

            return str(setuptools_scm.get_version(possible_root))
        else:
            try:
                from importlib.metadata import version as version_func
            except ModuleNotFoundError:
                from importlib_metadata import version as version_func  # type: ignore

            return str(version_func("portmod"))


_DIRS = AppDirs("portmod")


class Env:
    ALLOW_LOAD_ERROR = True
    DEBUG = False
    DATA_DIR = _DIRS.user_data_dir
    CONFIG_DIR = _DIRS.user_config_dir
    CACHE_DIR = _DIRS.user_cache_dir
    INTERACTIVE = True
    TESTING = False
    PREFIX_NAME: Optional[str] = None
    PREFIX_ARCH: Optional[str] = None
    REPOS: List["portmod.repo.Repo"]

    # The following variables are none if a prefix has not been selected
    class Prefix:
        REPOS: List["portmod.repo.Repo"]

        def __init__(self, prefix: str):
            from portmod.prefix import get_prefixes

            if prefix not in get_prefixes():
                raise Exception(l10n("invalid-prefix", prefix=prefix))

            self.ARCH = get_prefixes()[prefix]
            self.CONFIG_DIR = os.path.join(env.CONFIG_DIR, prefix)
            self.SET_DIR = os.path.join(self.CONFIG_DIR, "sets")
            self.CONFIG = os.path.join(self.CONFIG_DIR, "portmod.conf")

            self.ROOT = os.path.join(env.DATA_DIR, prefix)
            self.MODULES_DIR = os.path.join(self.ROOT, "modules")
            self.VARIABLE_DATA = os.path.join(self.ROOT, "var")
            self.VAR_SET_DIR = os.path.join(self.VARIABLE_DATA, "sets")
            self.INSTALLED_DB = os.path.join(self.VARIABLE_DATA, "db")
            self.NEWS_DIR = os.path.join(self.VARIABLE_DATA, "news")

            self.CACHE_DIR = os.path.join(env.CACHE_DIR, "prefix", prefix)
            self.PYBUILD_INSTALLED_CACHE = os.path.join(self.CACHE_DIR, "pybuild")
            self.CONFIG_PROTECT_DIR = os.path.join(self.CACHE_DIR, "cfg_protect")
            self.LOCAL_MODS = os.path.join(self.ROOT, "local")
            self.REPOS = []

        def __setattr__(self, name: str, value: Any):
            if isinstance(value, str):
                os.environ["PORTMOD_" + name] = value
            super().__setattr__(name, value)

    def prefix(self) -> Prefix:
        if self.PREFIX_NAME is not None:
            return self._prefix
        raise Exception("Internal error: Prefix has not been initialized!")

    def set_prefix(self, prefix: Optional[str]):
        from portmod.lock import has_prefix_exclusive_lock, has_prefix_lock
        from portmod.repos import get_repos

        if self.PREFIX_NAME:
            if has_prefix_exclusive_lock() or has_prefix_lock():
                raise Exception("Cannot change prefix while a lock is acquired!")

        if prefix:
            self._prefix = Env.Prefix(prefix)
            self.PREFIX_NAME = prefix
            self._prefix.REPOS = get_repos()
        else:
            self.PREFIX_NAME = None

    def __init__(self):
        self.PREFIX_FILE = os.path.join(self.DATA_DIR, "prefix")
        self.REPOS_DIR = os.path.join(self.DATA_DIR, "repos")
        self.PYBUILD_CACHE_DIR = os.path.join(self.CACHE_DIR, "pybuild")
        self.DOWNLOAD_DIR = os.path.join(self.CACHE_DIR, "downloads")
        self.GLOBAL_PORTMOD_CONFIG = os.path.join(self.CONFIG_DIR, "portmod.conf")
        self.REPOS_FILE = os.path.join(self.CONFIG_DIR, "repos.cfg")
        self.REPOS = []

        self.TMP_DIR = os.path.join(tempfile.gettempdir(), "portmod")
        self.PYBUILD_TMP_DIR = os.path.join(self.TMP_DIR, "pybuild")
        self.WARNINGS_DIR = os.path.join(self.PYBUILD_TMP_DIR, "messages")
        self.MESSAGES_DIR = os.path.join(self.PYBUILD_TMP_DIR, "warnings")

    def __setattr__(self, name: str, value: Any):
        if isinstance(value, str):
            os.environ["PORTMOD_" + name] = value
        super().__setattr__(name, value)

    def serialize(self):
        result = {}
        for key in dir(self):
            if not key.startswith("__") and not callable(getattr(self, key)):
                result[key] = getattr(self, key)
        return pickle.dumps(result)

    def deserialize(self, dump: bytes):
        result = pickle.loads(dump)
        for key in result:
            setattr(self, key, result[key])


env = Env()

# Inititialize env.REPOS
import portmod.repos  # noqa isort:skip



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3


import os

from portmod.prefix import add_prefix
from portmod.repo import get_repo
from portmod.repo.metadata import get_archs
from portmod.sync import sync
from portmodlib.l10n import l10n


def init(args):
    """Initializes a prefix"""
    add_prefix(args.prefix, args.arch)


def add_init_parser(subparsers, parents):
    parser = subparsers.add_parser("init", help=l10n("init-help"), parents=parents)
    parser.add_argument(
        "prefix", metavar=l10n("prefix-placeholder"), help=l10n("init-prefix-help")
    )
    try:
        meta_repo = get_repo("meta")
        if not os.path.exists(meta_repo.location):
            sync([meta_repo])
        parser.add_argument(
            "arch", help=l10n("init-arch-help"), choices=get_archs(meta_repo.location)
        )
    except Exception:
        parser.add_argument("arch", help=l10n("init-arch-help"))
    parser.set_defaults(func=init)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from contextlib import ContextDecorator
from typing import Optional

from .globals import env
from .lock import vdb_lock


class VDB(ContextDecorator):
    def __init__(self, commit_message: Optional[str] = None):
        # Slow import
        import git

        self.lock = vdb_lock(write=True)
        self.gitrepo = git.Repo.init(env.prefix().INSTALLED_DB)
        self.message = commit_message

    def __enter__(self):
        self.lock.__enter__()
        return self.gitrepo

    def __exit__(self, *exc):
        if self.message is not None:
            self.gitrepo.git.commit(m=self.message)
        self.lock.__exit__()
        return False



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from collections import defaultdict
from typing import Dict, List, Optional, Set

from portmod.functools import prefix_aware_cache
from portmod.globals import env
from portmod.repo import get_repo
from portmod.repo.metadata import get_masters
from portmod.repo.profiles import profile_exists, profile_parents
from portmodlib.atom import QualifiedAtom, atom_sat
from portmodlib.parsers.list import CommentedLine, read_list


@prefix_aware_cache
def get_masked(repo: Optional[str] = None) -> Dict[str, Dict[QualifiedAtom, List[str]]]:
    """
    Returns details about masked packages

    args:
        repo: Name of the repository in question, to be used outside the context of a prefix

    returns:
        A mapping of package Category-Package-Name strings to their comment and atom
        (noting that the atom can contain version specifiers, and the mapping is
        provided to facilitate efficient checks).
    """
    masked: Dict[str, Dict[QualifiedAtom, List[str]]] = defaultdict(dict)

    paths: List[str] = []

    if env.PREFIX_NAME:
        paths.extend(
            os.path.join(repo.location, "profiles") for repo in env.prefix().REPOS
        )
        paths.append(env.prefix().CONFIG_DIR)

        if profile_exists():
            paths.extend(profile_parents())
    elif repo:
        paths.append(get_repo(repo).location)
        for master in get_masters(repo):
            paths.append(master.location)

    for path in paths:
        if os.path.exists(os.path.join(path, "package.mask")):
            for line in read_list(os.path.join(path, "package.mask")):
                atom = QualifiedAtom(line)
                if isinstance(line, CommentedLine):
                    masked[atom.CPN][atom] = line.comment
                else:
                    masked[atom.CPN][atom] = []
    return masked


@prefix_aware_cache
def get_unmasked() -> Dict[str, Set[QualifiedAtom]]:
    """
    Returns a dictionary mapping Category-Package-Name strings to
    the precice atom which is unmasked
    """
    unmasked: Dict[str, Set[QualifiedAtom]] = defaultdict(set)

    path = env.prefix().CONFIG_DIR
    if os.path.exists(os.path.join(path, "package.unmask")):
        for line in read_list(os.path.join(path, "package.unmask")):
            atom = QualifiedAtom(line)
            unmasked[atom.CPN].add(atom)
    return unmasked


@prefix_aware_cache
def is_masked(atom: QualifiedAtom, repo: Optional[str] = None):
    if atom.CPN in get_masked(repo) and any(
        atom_sat(atom, masked) for masked in get_masked(repo)[atom.CPN]
    ):
        if atom.CPN not in get_unmasked() or not any(
            atom_sat(atom, unmasked) for unmasked in get_unmasked()[atom.CPN]
        ):
            return True

    return False



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3


class InputException(Exception):
    """Exception to indicate that the user has entered invalid input"""



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3
"""
Module building on behaviour from the builtin functools  module
"""

from functools import lru_cache, wraps
from typing import Any, Callable, Optional, TypeVar, cast

from .globals import env

F = TypeVar("F", bound=Callable[..., Any])


def prefix_aware_cache(func: F) -> F:
    """
    A variant of functools.lru_cache which treats the prefix as if it were an argument
    """

    @wraps(func)
    @lru_cache(maxsize=None)
    def inner(prefix: Optional[str], *args, **kwargs):
        return func(*args, **kwargs)

    @wraps(func)
    def prefix_wrapper(*args, **kwargs):
        return inner(env.PREFIX_NAME, *args, **kwargs)

    prefix_wrapper.cache_clear = inner.cache_clear  # type: ignore
    return cast(F, prefix_wrapper)


_INSTALL_CACHE_FUNCS = []


def install_cache(func: F) -> F:
    """
    A variant of functools.lru_cache which treats the prefix as if it were an argument
    """
    global _INSTALL_CACHE_FUNCS

    @wraps(func)
    @prefix_aware_cache
    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    _INSTALL_CACHE_FUNCS.append(inner)

    return cast(F, inner)


def clear_install_cache():
    for func in _INSTALL_CACHE_FUNCS:
        func.cache_clear()  # type: ignore



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Dependency resolution integration tests
"""

import pytest

from portmod._deps import DepError
from portmod.config.use import get_use
from portmod.loader import load_all_installed, load_installed_pkg
from portmod.merge import configure
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("base")  # Like test, but has no system mods
    yield dictionary
    tear_down_env()


def test_deps(setup):
    """
    Tests that dependencies resolve correctly
    """
    # test3 test4 and test5 should resolve correctly if
    # the foo flag is disabled for mod 3
    configure(["test/test3", "test/test5"], no_confirm=True)
    mod3 = load_installed_pkg(Atom("test/test3"))
    mod4 = load_installed_pkg(Atom("test/test4"))
    mod5 = load_installed_pkg(Atom("test/test5"))
    installed = set(load_all_installed())
    assert mod3 in installed
    assert mod4 in installed
    assert mod5 in installed
    assert "foo" not in get_use(mod3)[0]

    # test3, test4, test5, and test6 should not resolve correctly
    with pytest.raises(DepError):
        configure(["test/test3", "test/test5", "test/test6"], no_confirm=True)


def test_unsolveable(setup):
    """Tests that mods that cannot be resolved fail to resolve"""

    with pytest.raises(DepError):
        configure(["test/test-useless"], no_confirm=True)


def test_weight(setup):
    """Tests that mods which are not selected are not pulled in by the dependency calculation"""

    # Note test7 will cause other mods to be pulled into the formula due to conditionals,
    # but by default shouldn't depend on anything
    configure(["test/test7"], no_confirm=True)
    assert len(list(load_all_installed())) == 1
    assert all(mod.PN != "test" for mod in load_all_installed())


def test_usedep_cli(setup):
    """Tests that use dependencies are satisfied when passed on the command line"""
    configure(["test/test4[foo,bar,-baz]"], no_confirm=True)
    mod = load_installed_pkg(Atom("test/test4"))
    assert mod
    assert mod.INSTALLED_USE >= {"foo", "bar"}
    assert "baz" not in mod.INSTALLED_USE



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Functions for interacting with the OpenMW VFS
"""

import os
import shutil
from tempfile import gettempdir
from typing import Dict, List

from portmodlib.fs import ci_exists
from portmodlib.l10n import l10n

from .archives import extract_archive_file, list_archive
from .functools import vfs_cache
from .globals import root
from .parsers.list import read_list


@vfs_cache
def find_file(name: str) -> str:
    """
    Locates the path of a file within the OpenMW virtual file system

    args:
        name: The relative path within the VFS to search for

    returns:
        The absolute path of the file
    """
    # FIXME: This should respect the CASE_INSENSITIVE_FILES setting
    for directory in reversed(get_vfs_dirs()):
        path = ci_exists(os.path.join(directory, name))
        if path:
            return path

    for archive in reversed(get_vfs_archives()):
        contents = list_archive(archive)
        for file in contents:
            if os.path.normpath(file).lower() == os.path.normpath(name).lower():
                return extract_archive_file_to_tmp(archive, file)

    raise FileNotFoundError(name)


@vfs_cache
def list_dir(name: str) -> List[str]:
    """
    Locates all path of files matching the given pattern within the OpenMW
    virtual file system

    args:
        name: The relative path of the directory within the VFS
    returns:
        A list of files contained within the directory
    """
    files: Dict[str, str] = {}
    normalized = os.path.normpath(name).lower()

    # FIXME: This should respect the CASE_INSENSITIVE_FILES setting

    for directory in reversed(get_vfs_dirs()):
        path = ci_exists(os.path.join(directory, normalized))
        if path:
            for file in os.listdir(path):
                if file.lower() not in files:
                    files[file.lower()] = file

    for archive in reversed(get_vfs_archives()):
        contents = list_archive(archive)
        for file in contents:
            if (
                os.path.commonpath([normalized, os.path.normpath(file).lower()])
                == normalized
            ):
                suffix = os.path.relpath(os.path.normpath(file).lower(), normalized)
                component, _, _ = suffix.partition(os.sep)
                files[component] = component

    return sorted(files.values())


def _cleanup_tmp_archive_dir():
    path = os.path.join(gettempdir(), ".archive_files")
    if os.path.exists(path):
        shutil.rmtree(path)


def extract_archive_file_to_tmp(archive: str, file: str) -> str:
    """Extracts the given file from the archive and places it in a temprorary directory"""
    temp = gettempdir()
    output_dir = os.path.join(
        temp, ".archive_files", os.path.basename(archive), os.path.dirname(file)
    )
    os.makedirs(output_dir, exist_ok=True)
    result_file = os.path.join(output_dir, os.path.basename(file))
    extract_archive_file(archive, file, output_dir)
    if not os.path.exists(result_file):
        raise Exception(l10n("archive-extraction-failed", file=file, dest=result_file))
    return result_file


@vfs_cache
def get_vfs_dirs() -> List[str]:
    """Returns an ordered list of the VFS directories, in reverse order of priority"""
    dirs: List[str] = read_list(os.path.join(root(), "var", "vfs"))
    return dirs


@vfs_cache
def get_vfs_archives() -> List[str]:
    """Returns an ordered list of the VFS directories, in reverse order of priority"""
    archives: List[str] = read_list(os.path.join(root(), "var", "vfs-archives"))
    return archives



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""A module for parsing package updates"""

import os
from typing import Any, Dict, Optional

from portmod.globals import env
from portmod.repos import Repo

_UPDATES: Optional[Dict[str, Dict[str, Any]]] = None


def get_moved(repo: Repo) -> Dict[str, str]:
    global _UPDATES
    if _UPDATES is None:
        _UPDATES = {}
        for repo in env.REPOS:
            parse_updates(repo)

    assert _UPDATES is not None
    return _UPDATES.get(repo.name, {}).get("move", {})


def parse_updates(repo: Repo):
    global _UPDATES
    path = os.path.join(repo.location, "profiles", "updates")
    repo_updates: Dict[str, Any] = {}

    def update_gt(update_filename: str, other: str):
        if not other:
            return True

        if not update_filename:
            return False

        update_parts = reversed(update_filename.split("-"))
        other_parts = reversed(other.split("-"))
        for component1, component2 in zip(update_parts, other_parts):
            if component1 > component2:
                return True
        return False

    def parse_file(filename: str):
        result = {}
        with open(filename, "r") as file:
            for line in file.readlines():
                command, _, arguments = line.partition(" ")
                if command == "move":
                    source, target = arguments.split(" ")
                    source = source.strip()
                    target = target.strip()
                    if "move" not in result:
                        result["move"] = {source: target}
                    else:
                        result["move"][source] = target
        return result

    newest = repo_updates.get("newest", "")
    if os.path.exists(path):
        for filename in os.listdir(path):
            if update_gt(filename, repo_updates.get("newest", "")):
                if update_gt(filename, newest):
                    newest = filename
                repo_updates.update(parse_file(os.path.join(path, filename)))

    assert _UPDATES is not None
    _UPDATES[repo.name] = repo_updates



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Quality assurance for the mod repo
"""

import argparse
import glob
import os
import re
import sys
import traceback
from logging import debug, error
from typing import Callable

from portmod.globals import env
from portmod.loader import SandboxedError, load_file
from portmod.news import validate_news
from portmod.repo import Repo, get_repo_name, get_repo_root
from portmod.repo.loader import get_atom_from_path
from portmod.repo.metadata import (
    get_categories,
    get_use_flag_atom_aliases,
    license_exists,
)
from portmodlib.atom import Atom, InvalidAtom, UnqualifiedAtom
from portmodlib.l10n import l10n
from portmodlib.log import add_logging_arguments, init_logger
from portmodlib.parsers.list import read_list
from portmodlib.portmod import (
    parse_category_metadata,
    parse_groups,
    parse_yaml_dict,
    parse_yaml_dict_dict,
)

from .pybuild import pybuild_manifest, pybuild_validate


def scan_package_dir(path: str, err: Callable[[str], None]):
    cpn = os.path.join(os.path.basename(os.path.dirname(path)), os.path.basename(path))
    if Atom(os.path.basename(cpn)).PV is not None:
        err(f"Package name {cpn} must not end in a version")

    for file in glob.glob(os.path.join(path, "*.pybuild")):
        debug(f"Scanning {file}")
        dir_name = os.path.basename(path)
        file_name = Atom(os.path.splitext(os.path.basename(file))[0]).PN
        if dir_name != file_name:
            err(
                f"The package name in filename {file} should match its parent directory's name!"
            )

        try:
            pybuild_validate(file)
        except SandboxedError as e:
            err(f"{e}")
        except Exception as e:
            traceback.print_exc()
            err(f"{e}")


def scan_category_metadata(path: str, err: Callable[[str], None]):
    # Note: Package metadata is already validated as part of pybuild_validate
    try:
        parse_category_metadata(path)
    except Exception as e:
        traceback.print_exc()
        err("{}".format(e))


def scan_category(path: str, err: Callable[[str], None]):
    for directory in glob.glob(os.path.join(path, "*")):
        if os.path.isdir(directory) and any(
            file.lower().endswith(".pybuild") for file in os.listdir(directory)
        ):
            scan_package_dir(directory, err)
    metadata_path = os.path.join(path, "metadata.yaml")
    if os.path.exists(metadata_path):
        scan_category_metadata(metadata_path, err)


def scan_arch_list(repo_root: str, err: Callable[[str], None]):
    # Check profiles/arch.list
    path = os.path.join(repo_root, "profiles", "arch.list")
    if os.path.exists(path):
        archs = read_list(path)
        for arch in archs:
            if " " in arch:
                err(
                    f'arch.list: in entry "{arch}". '
                    "Architectures cannot contain spaces"
                )


def scan_categories(repo_root: str, err: Callable[[str], None]):
    # Check profiles/categories
    path = os.path.join(repo_root, "profiles", "categories")
    if os.path.exists(path):
        lines = read_list(path)
        for category in lines:
            if " " in category:
                err(
                    f'categories.list: in category "{category}". '
                    "Categories cannot contain spaces"
                )


def scan_groups(repo_root: str, err: Callable[[str], None]):
    # Check metadata/groups.yaml
    path = os.path.join(repo_root, "metadata", "groups.yaml")
    if os.path.exists(path):
        parse_groups(path)


def scan_license_groups(repo_root: str, err: Callable[[str], None]):
    # Check metadata/license_groups.yaml
    # All licenses should exist in licenses/LICENSE_NAME
    path = os.path.join(repo_root, "profiles", "license_groups.yaml")
    if os.path.exists(path):
        license_groups = parse_yaml_dict(path)
        for key, value in license_groups.items():
            if value is not None:
                for license in value.split():
                    if not license_exists(repo_root, license) and not (
                        license.startswith("@")
                    ):
                        err(
                            f'license_groups.yaml: License "{license}" in group {key} '
                            "does not exist in licenses directory"
                        )


def scan_repo_name(repo_root: str, err: Callable[[str], None]):
    # Check profiles/repo_name
    path = os.path.join(repo_root, "profiles", "repo_name")
    if os.path.exists(path):
        lines = read_list(path)
        if len(lines) == 0:
            err("repo_name: profiles/repo_name cannot be empty")
        elif len(lines) > 1:
            err(
                "repo_name: Extra lines detected. "
                "File must contain just the repo name."
            )
        elif " " in lines[0]:
            err("repo_name: Repo name must not contain spaces.")


def scan_use(repo_root: str, err: Callable[[str], None]):
    # Check profiles/use.yaml
    path = os.path.join(repo_root, "profiles", "use.yaml")
    if os.path.exists(path):
        flags = parse_yaml_dict(path)
        for desc in flags.values():
            if not isinstance(desc, str):
                err(f'use.yaml: Description "{desc}" is not a string')


def scan_profiles(repo_root: str, err: Callable[[str], None]):
    # Check profiles/profiles.yaml
    path = os.path.join(repo_root, "profiles", "profiles.yaml")
    archs = read_list(os.path.join(repo_root, "profiles", "arch.list"))
    if os.path.exists(path):
        keywords = parse_yaml_dict_dict(path)
        for keyword, profiles in keywords.items():
            if keyword not in archs:
                err(
                    f"profiles.yaml: keyword {keyword} " "was not declared in arch.list"
                )
            for profile in profiles:
                if not isinstance(profile, str):
                    err('profiles.yaml: Profile "{profile}" is not a string')
                path = os.path.join(repo_root, "profiles", profile)
                if not os.path.exists(path):
                    err(f"profiles.yaml: Profile {path} does not exist")


def scan_use_expand(filename: str, err: Callable[[str], None]):
    entries = parse_yaml_dict(filename)
    for entry in dict(entries):
        if not re.match("[A-Za-z0-9][A-Za-z0-9+_-]*", entry):
            err(f"USE_EXPAND flag {entry} in {filename} contains invalid characters")


def scan_root(repo_root: str, err: Callable[[str], None]):
    # Run pybuild validate on every pybuild in repo
    for category in get_categories(repo_root):
        scan_category(os.path.join(repo_root, category), err)

    # Check files in metadata and profiles.
    # These may not exist, as they might be inherited from another repo instead
    scan_arch_list(repo_root, err)
    scan_categories(repo_root, err)
    scan_groups(repo_root, err)
    scan_license_groups(repo_root, err)
    scan_repo_name(repo_root, err)
    scan_use(repo_root, err)
    scan_profiles(repo_root, err)
    for filename in glob.glob(os.path.join(repo_root, "profiles", "desc", "*.yaml")):
        scan_use_expand(filename, err)
    # Check news
    validate_news(repo_root, err)
    scan_use_alias(repo_root, err)


def scan_use_alias(repo_root: str, err: Callable[[str], None]):
    # Check profiles/use.alias.yaml
    try:
        aliases = get_use_flag_atom_aliases(repo_root)
    except (RuntimeError, UnqualifiedAtom, InvalidAtom) as _e:
        err(str(_e))
    for flag, alias in aliases.items():
        # Check that alias is a valid package in the repository
        packages = glob.glob(
            os.path.join(repo_root, alias.C, alias.PN, alias.PN + "-*.pybuild")
        )
        if not packages:
            err(
                f"Use flag alias package {alias} for flag {flag} "
                f"does not exist in this repository"
            )

        # If alias has a use flag requirement, it must be a valid use flag for at least
        # one version of that package
        if alias.USE:
            if not any(
                alias.USE <= set(load_file(package).IUSE_EFFECTIVE)
                for package in packages
            ):
                err(
                    "No package has the required use flag dependencies "
                    f"for use flag {flag}'s alias {alias}"
                )


def scan_file(filename: str, repo_root: str, err: Callable[[str], None]):
    _, ext = os.path.splitext(filename)
    relative = os.path.normpath(os.path.relpath(filename, start=repo_root))
    if ext.lower() == ".pybuild":
        scan_package_dir(os.path.dirname(filename), err)
    else:
        news_dir = os.path.join("metadata", "news")
        if relative == os.path.join("profiles", "arch.list"):
            scan_arch_list(repo_root, err)
        elif relative == os.path.join("profiles", "categories"):
            scan_categories(repo_root, err)
        elif relative == os.path.join("metadata", "groups.yaml"):
            scan_groups(repo_root, err)
        elif relative == os.path.join("profiles", "license_groups.yaml"):
            scan_license_groups(repo_root, err)
        elif relative == os.path.join("profiles", "repo_name"):
            scan_repo_name(repo_root, err)
        elif relative == os.path.join("profiles", "use.yaml"):
            scan_use(repo_root, err)
        elif relative == os.path.join("profiles", "use.alias.yaml"):
            scan_use_alias(repo_root, err)
        elif os.path.dirname(relative) == os.path.join(
            "profiles", "desc"
        ) and relative.endswith(".yaml"):
            scan_use_expand(filename, err)
        elif os.path.commonprefix([relative, news_dir]) == news_dir:
            validate_news(repo_root, err)
        elif os.path.basename(filename) == "metadata.yaml":
            path, _ = os.path.split(relative)
            if os.path.split(path)[0] is None:
                scan_category_metadata(filename, err)
            else:
                scan_package_dir(os.path.dirname(filename), err)


def scan_commit(commit, err):
    import git

    _git = git.Git()

    files = _git.show(commit, name_only=True, oneline=True).splitlines()[1:]
    message = _git.log("HEAD", format="%B", n=1)
    header_line = message.splitlines()[0]
    packages_modified = [file for file in files if file.endswith(".pybuild")]
    if len(packages_modified) == 1:
        atom = get_atom_from_path(packages_modified[0]).CPN
        if not message.startswith(atom + ":"):
            err(f'Commit "{header_line}" should start with "{atom}: <short desc>"')


def commit_message(args, repo_root: str, err: Callable[[str], None]):
    import git

    gitrepo = git.Repo.init(repo_root)

    initial_message = None
    message = ""
    if args.initial_message:
        with open(args.initial_message) as file:
            initial_message = file.read()

    changes = gitrepo.head.commit.diff(git.Diffable.Index)

    pybuild_diffs = [diff for diff in changes if diff.b_path.endswith(".pybuild")]

    if len(pybuild_diffs) == 1:
        diff = pybuild_diffs[0]

        if diff.a_path.endswith(".pybuild"):
            if diff.a_path:
                old = get_atom_from_path(diff.a_path)
            if diff.b_path:
                new = get_atom_from_path(diff.b_path)

            if diff.deleted_file:
                message = f"{old.CPN}: Removed version {old.PV}"
                if initial_message:
                    message += f"\n\n{initial_message}"
            elif diff.new_file:
                message = f"{new.CPN}: Added version {new.PV}"
                if initial_message:
                    message += f"\n\n{initial_message}"
            elif diff.renamed_file and old.PV != new.PV:
                message = f"{new.CPN}: Updated to version {new.PV}"
                if initial_message:
                    message += f"\n\n{initial_message}"
            else:
                # Either a change to the package without bump, or just a revision bump.
                # We can't autogenerate a meaningful message
                if initial_message:
                    if initial_message.startswith(new.CPN + ":"):
                        message = initial_message
                    else:
                        message = f"{new.CPN}: {initial_message}"
                else:
                    message = f"{new.CPN}: "

        if args.initial_message:
            with open(args.initial_message, "w") as file:
                file.write(message)
        else:
            print(message)


def scan(args, repo_root: str, err: Callable[[str], None]):
    if args.diff:
        import git

        for file in git.Git().diff(args.diff, name_only=True).splitlines():
            scan_file(os.path.join(repo_root, file), repo_root, err)

        for commit in git.Git().log("HEAD", "^" + args.diff, pretty="%H").splitlines():
            scan_commit(commit, err)

    else:
        for root in args.paths or [os.getcwd()]:
            if os.path.exists(os.path.join(root, "profiles", "repo_name")):
                scan_root(root, err)
            elif os.path.isdir(root):
                if glob.glob(os.path.join(root, "*", "*.pybuild")):
                    scan_category(root, err)
                elif glob.glob(os.path.join(root, "*.pybuild")):
                    scan_package_dir(root, err)
                else:
                    # Try to scan all files in directory tree
                    for path_root, _, filenames in os.walk(root):
                        for filename in filenames:
                            scan_file(os.path.join(path_root, filename), repo_root, err)
            elif os.path.isfile(root):
                scan_file(root, repo_root, err)


def manifest(args, _repo_root: str, err: Callable[[str], None]):
    def try_manifest(file: str):
        try:
            pybuild_manifest(file)
        except Exception as e:
            traceback.print_exc()
            err(f"{e}")

    for root in args.paths or [os.getcwd()]:
        if os.path.isdir(root):
            # Run pybuild manifest on every pybuild in repo
            for file in glob.iglob(
                os.path.join(root, "**", "*.pybuild"), recursive=True
            ):
                try_manifest(file)
        else:
            _, ext = os.path.splitext(root)
            if ext.lower() == ".pybuild":
                try_manifest(root)
            else:
                err(f"{root} is not a pybuild file!")


def main():
    """
    Main function for the inquisitor executable
    """

    common = argparse.ArgumentParser(add_help=False)
    add_logging_arguments(common)
    common.add_argument("--debug", help=l10n("merge-debug-help"), action="store_true")

    parser = argparse.ArgumentParser(
        description="Quality assurance program for the package repository",
        parents=[common],
    )
    subparsers = parser.add_subparsers()
    manifest_parser = subparsers.add_parser(
        "manifest", help="Produces Manifest files", parents=[common]
    )
    scan_parser = subparsers.add_parser(
        "scan", help="QA Checks package repositories", parents=[common]
    )
    commit_msg_parser = subparsers.add_parser(
        "commit-msg",
        help="Produces a commit message using the working directory. "
        "Designed to be used as a git commit-msg hook",
        parents=[common],
    )
    commit_msg_parser.add_argument(
        "initial_message",
        help="Path to a file containing a user-supplied message to start from",
        nargs="?",
    )
    commit_msg_parser.add_argument(
        "--paths",
        help="Location of the repository to process. "
        "If omitted, the current working directory will be used",
        nargs="?",
    )
    scan_parser.add_argument(
        "--diff",
        nargs="?",
        help="Scan files changed since the given git target (branch, commit, etc.)",
    )
    manifest_parser.add_argument(
        "paths",
        metavar="PATH",
        help="scope to process. If not provided defaults to the current working directory",
        nargs="*",
    )
    scan_parser.add_argument(
        "paths",
        metavar="PATH",
        help="scope to process. If not provided defaults to the current working directory",
        nargs="*",
    )

    scan_parser.set_defaults(func=scan)
    manifest_parser.set_defaults(func=manifest)
    commit_msg_parser.set_defaults(func=commit_message)

    args = parser.parse_args()
    init_logger(args)

    if hasattr(args, "paths") and args.paths:
        repo_root = get_repo_root(args.paths[0])
    else:
        repo_root = get_repo_root(os.getcwd())

    has_errored = False
    env.ALLOW_LOAD_ERROR = False

    def err(string: str):
        nonlocal has_errored
        error(string)
        has_errored = True

    if repo_root is None:
        err(
            "Cannot find repository for the current directory. "
            "Please run from within the repository you wish to inspect"
        )
        sys.exit(1)

    # Register repo in case it's not already in repos.cfg
    real_root = os.path.realpath(repo_root)
    if not any([real_root == os.path.realpath(repo.location) for repo in env.REPOS]):
        sys.path.insert(0, os.path.join(repo_root))
        env.REPOS.insert(0, Repo(get_repo_name(repo_root), repo_root))

    if args.debug:
        env.DEBUG = True

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(2)

    args.func(args, repo_root, err)

    if has_errored:
        sys.exit(1)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Interface for searching for packages"""

from portmod.query import display_search_results, query
from portmodlib.l10n import l10n


def search_main(args):
    if args.description:
        pkgs = query(
            ["NAME", "ATOM", "DESC"],
            " ".join(args.query),
            strip=True,
            squelch_sep=True,
            insensitive=True,
        )
    else:
        pkgs = query(
            ["NAME", "ATOM"],
            " ".join(args.query),
            strip=True,
            squelch_sep=True,
            insensitive=True,
        )

    display_search_results(pkgs)


def add_search_parser(subparsers, parents):
    parser = subparsers.add_parser("search", help=l10n("search-help"), parents=parents)
    parser.add_argument(
        "query",
        metavar=l10n("query-placeholder"),
        help=l10n("search-query-help"),
        nargs="+",
    )
    parser.add_argument(
        "-d", "--description", help=l10n("searchdesc-help"), action="store_true"
    )
    parser.set_defaults(func=search_main)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
import traceback
from logging import error

from portmod.download import download_mod, mirrorable
from portmod.loader import load_all
from portmodlib.l10n import l10n


def add_mirror_parser(subparsers, parents):
    parser = subparsers.add_parser("mirror", help=l10n("mirror-help"), parents=parents)
    parser.add_argument(
        "mirror", metavar=l10n("directory-placeholder"), help=l10n("mirror-dir-help")
    )

    def mirror_main(args):
        if args.mirror:
            os.makedirs(args.mirror, exist_ok=True)
            for mod in load_all():
                if mirrorable(mod):
                    try:
                        for source in download_mod(mod, True):
                            path = os.path.join(args.mirror, source.name)
                            if not os.path.exists(path):
                                print(l10n("copying-file", src=source.path, dest=path))
                                shutil.copy(source.path, path)
                    except Exception as e:
                        traceback.print_exc()
                        error(f"{e}")
        else:
            parser.print_help()

    parser.set_defaults(func=mirror_main)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
import sys

import pytest

import pybuild
from portmod._cli.pybuild import pybuild_manifest
from portmod.cache import clear_cache_for_path
from portmod.globals import env
from portmod.loader import SandboxedError, _sandbox_execute_pybuild, load_file
from portmod.package import install_pkg, src_prepare, src_unpack
from portmod.perms import Permissions
from portmod.repo import Repo
from portmod.repo.loader import __safe_load_file
from portmod.win32 import get_personal

from .env import TEST_REPO_DIR, setup_env, tear_down_env

TMP_REPO = os.path.join(os.path.dirname(env.TMP_DIR), "not-portmod")
TMP_VERSION = 0
TMP_FILE = os.path.join(TMP_REPO, "test", "test", f"test-{TMP_VERSION}.pybuild")

# Windows restricted files are greatly limited due to limitations of Sandboxie,
# but we should test them anyway
if sys.platform == "win32":
    BLOCKED_FILE = os.path.join(get_personal(), "portmod-not-allowed")
else:
    BLOCKED_FILE = os.path.join(os.path.dirname(env.TMP_DIR), "portmod-not-allowed")
env.ALLOW_LOAD_ERROR = False


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    env.REPOS.append(Repo("test", TMP_REPO))
    yield dictionary
    tear_down_env()
    shutil.rmtree(TMP_REPO)
    shutil.rmtree(env.TMP_DIR)


def create_pybuild(filestring, manifest: bool = False):
    global TMP_VERSION, TMP_FILE
    TMP_VERSION += 1
    TMP_FILE = os.path.join(TMP_REPO, "test", "test", f"test-{TMP_VERSION}.pybuild")
    if (
        "{TMP_FILE}" in filestring
        or "{BLOCKED_FILE}" in filestring
        or "{TMP_DIR}" in filestring
    ):
        filestring = filestring.format(
            TMP_FILE=TMP_FILE, BLOCKED_FILE=BLOCKED_FILE, TMP_DIR=env.TMP_DIR
        )
    os.makedirs(env.TMP_DIR, exist_ok=True)
    os.makedirs(os.path.dirname(TMP_FILE), exist_ok=True)
    os.makedirs(os.path.join(TMP_REPO, "profiles"), exist_ok=True)
    clear_cache_for_path(TMP_FILE)

    if sys.platform == "win32":
        filestring = filestring.replace("\\", "\\\\")

    with open(TMP_FILE, "w") as file:
        file.write(filestring)

    with open(BLOCKED_FILE, "w") as file:
        print(file=file)

    with open(os.path.join(TMP_REPO, "profiles", "repo_name"), "w") as file:
        file.write("test")

    with open(os.path.join(TMP_REPO, "profiles", "categories"), "w") as file:
        file.write("test")

    if manifest:
        pybuild_manifest(TMP_FILE)

    return TMP_FILE


def full_load_file(file: str):
    return __safe_load_file(file)


def test_safe():
    """Tests that a simple safe pybuild loads correctly"""
    file = """
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    create_pybuild(file)
    __safe_load_file(TMP_FILE)


def test_write_globalscope():
    """Tests that writing to files outside TMP_DIR is forbidden in the global scope"""
    file = """
import os
from pybuild import Pybuild1

os.remove("{TMP_FILE}")

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    with pytest.raises(RuntimeError):
        create_pybuild(file)
        full_load_file(TMP_FILE)
    assert os.path.exists(TMP_FILE)


def test_write_globalscope_2():
    """Tests that writing to files within TMP_DIR is forbidden in the global scope"""
    file = """
import os
from pybuild import Pybuild1

os.remove("{TMP_DIR}")

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    with pytest.raises(RuntimeError):
        create_pybuild(file)
        full_load_file(TMP_FILE)
    assert os.path.exists(TMP_FILE)


def test_read_globalscope():
    """Tests that reading from files outside TMP_DIR is forbidden in the global scope"""
    file = """
import os
from pybuild import Pybuild1

assert os.path.exists("{BLOCKED_FILE}")

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    with pytest.raises(RuntimeError):
        create_pybuild(file)
        full_load_file(TMP_FILE)
    assert os.path.exists(TMP_FILE)


def test_read_src_unpack():
    """Tests that reading from files outside TMP_DIR is forbidden within src_unpack"""
    file = """
import os
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        assert os.path.exists("{BLOCKED_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_unpack(mod, env.TMP_DIR)


@pytest.mark.xfail(
    sys.platform == "win32",
    reason="Sandboxie's permissions don't allow nesting, so we have to use the "
    "default redirected writes, which don't raise exceptions",
)
def test_write_src_unpack():
    """Tests that writing to files outside TMP_DIR is forbidden within src_unpack"""
    file = """
import os
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        os.remove("{TMP_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_unpack(mod, env.TMP_DIR)
    assert os.path.exists(TMP_FILE)


def test_read_can_update_live():
    """
    Tests that reading from files outside TMP_DIR is forbidden within can_update_live
    """
    file = """
import os
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def can_update_live(self):
        assert os.path.exists("{BLOCKED_FILE}")
        return True
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        _sandbox_execute_pybuild(mod.FILE, "can-update-live", Permissions(network=True))


@pytest.mark.xfail(
    sys.platform == "win32",
    reason="Sandboxie's permissions don't allow nesting, so we have to use the "
    "default redirected writes, which don't raise exceptions",
)
def test_write_can_update_live():
    """
    Tests that writing to files outside TMP_DIR is forbidden within can_update_live
    """
    file = """
import os
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def can_update_live(self):
        os.remove("{TMP_FILE}")
        return True
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        _sandbox_execute_pybuild(mod.FILE, "can-update-live", Permissions(network=True))
    assert os.path.exists(TMP_FILE)


def test_read_src_prepare():
    """Tests that reading from files outside TMP_DIR is allowed within src_prepare"""
    file = """
import os
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        assert os.path.exists("{BLOCKED_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)


@pytest.mark.xfail(
    sys.platform == "win32",
    reason="Sandboxie's permissions don't allow nesting, so we have to use the "
    "default redirected writes, which don't raise exceptions",
)
def test_write_src_prepare():
    """Tests that writing to files outside TMP_DIR is not allowed within src_prepare"""
    file = """
import os
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        os.remove("{TMP_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_prepare(mod, env.TMP_DIR)
    assert os.path.exists(TMP_FILE)


def test_formatstr():
    """Tests that string.format is banned"""
    file = """
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="{}".format("Test")
    LICENSE="GPL-3"
"""
    with pytest.raises(NotImplementedError):
        create_pybuild(file)
        __safe_load_file(TMP_FILE)


def test_module_write():
    """Tests that module changes do not propagate from within pybuilds"""
    file = """
import sys
from pybuild import Pybuild1

sys.platform = "foo"

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    old_platform = sys.platform
    create_pybuild(file)
    load_file(TMP_FILE)
    assert sys.platform == old_platform


def test_module_write_2():
    """Tests that module subattribute changes do not propagate from within pybuilds"""
    file = """
import pybuild
from pybuild import Pybuild1

pybuild.Pybuild1.src_install = "foo"

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    create_pybuild(file)
    load_file(TMP_FILE)
    assert pybuild.Pybuild1.src_install != "foo"


def test_module_unsafe_import():
    """Tests that modules cannot be imported indirectly"""
    file = """
from typing import sys
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    with pytest.raises(SyntaxError):
        create_pybuild(file)
        __safe_load_file(TMP_FILE)


@pytest.mark.xfail(reason="Portability feature not implemented")
def test_module_unsafe_import_2():
    """Tests that modules cannot be accessed indirectly"""
    file = """
import shutil
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
    def src_prepare(self):
        assert shutil.stat is not None
"""
    with pytest.raises(SyntaxError):
        create_pybuild(file)
        pkg = load_file(TMP_FILE)
        src_prepare(pkg, env.TMP_DIR)


def test_underscore():
    """Tests that we can't access names beginning with underscores"""
    file = """
import pybuild
from pybuild import Pybuild1

pybuild.__globals__

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""
    with pytest.raises(SyntaxError):
        create_pybuild(file)
        __safe_load_file(TMP_FILE)


def test_getattr():
    """Tests that we can't use getattr to access names beginning with underscores"""
    file = """
import os
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        getattr(os.path.join, "__globals__")
"""
    with pytest.raises(SandboxedError):
        create_pybuild(file)
        pkg = load_file(TMP_FILE)
        src_prepare(pkg, env.TMP_DIR)


def test_execute():
    """Tests that we can execute files properly even if we fiddle with platform"""
    file = """
from pybuild import Pybuild1
import sys


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        # If this affected the scope of execute, this would
        # raise an unsupported platform exception
        origplatform = sys.platform
        sys.platform = "foo"

        if origplatform == "win32":
            self.execute("dir")
        else:
            self.execute("ls")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)


def test_execute_src_prepare_write():
    """Tests that we can't modify files through execute in src_prepare"""
    file = """
from pybuild import Pybuild1
import sys


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        # If this affected the scope of execute, this would
        # raise an unsupported platform exception
        origplatform = sys.platform
        sys.platform = "foo"
        if origplatform == "win32":
            self.execute('cmd /c "del {TMP_FILE}"')
        else:
            self.execute("rm {TMP_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    if sys.platform == "win32":
        # Sandboxie won't cause an error when deleting the file fails
        src_prepare(mod, env.TMP_DIR)
    else:
        with pytest.raises(SandboxedError):
            src_prepare(mod, env.TMP_DIR)
    assert os.path.exists(TMP_FILE)


def test_execute_src_prepare_read():
    """Tests that we can read files through execute in src_prepare"""
    file = """
from pybuild import Pybuild1
import sys


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        # If this affected the scope of execute, this would
        # raise an unsupported platform exception
        origplatform = sys.platform
        sys.platform = "foo"

        if origplatform == "win32":
            self.execute('dir "{BLOCKED_FILE}"')
        else:
            self.execute("ls {BLOCKED_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)


def test_execute_src_unpack_read():
    """Tests that we can't read files through execute in src_unpack"""
    file = """
from pybuild import Pybuild1
import sys


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        # If this affected the scope of execute, this would
        # raise an unsupported platform exception
        origplatform = sys.platform
        sys.platform = "foo"
        if origplatform == "win32":
            self.execute('dir "{BLOCKED_FILE}"')
        else:
            self.execute("ls {BLOCKED_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_unpack(mod, env.TMP_DIR)


def test_execute_network_permissions():
    """Tests that network permissions work in src_unpack but not src_prepare"""
    file = """
from pybuild import Pybuild1
import sys

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        if sys.platform == "win32":
            self.execute('ping /n 1 1.1.1.1')
        else:
            self.execute('curl 1.1.1.1')

    def src_prepare(self):
        if sys.platform == "win32":
            self.execute('ping /n 1 1.1.1.1')
        else:
            self.execute('curl 1.1.1.1')
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_unpack(mod, env.TMP_DIR)
    with pytest.raises(SandboxedError):
        src_prepare(mod, env.TMP_DIR)


def test_execute_permissions_bleed():
    """Tests that network permissions from src_unpack don't bleed into src_prepare"""
    file = """
from pybuild import Pybuild1
import sys

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        if sys.platform == "win32":
            self.execute('dir "{TMP_DIR}"')
        else:
            self.execute("ls {TMP_DIR}")

    def src_prepare(self):
        if sys.platform == "win32":
            self.execute('ping /n 1 gitlab.com')
        else:
            self.execute('curl https://gitlab.com')
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_unpack(mod, env.TMP_DIR)
    with pytest.raises(SandboxedError):
        src_prepare(mod, env.TMP_DIR)


def test_default():
    """Tests that wrapped functions with default arguments work properly"""
    file = """
from pybuild import Pybuild1
import os


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        os.listdir()
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)


def test_execute_escape():
    """Tests that you can't escape from execute"""
    file = """
from pybuild import Pybuild1
import sys


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        if sys.platform == "win32":
            self.execute('& del "{TMP_FILE}"')
        else:
            self.execute("; rm {TMP_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_prepare(mod, env.TMP_DIR)
    assert os.path.exists(TMP_FILE)


def test_execute_escape_2():
    """Tests that you can't escape from execute"""
    file = """
from pybuild import Pybuild1
import sys


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        if sys.platform == "win32":
            self.execute('"& cmd /c "del {TMP_FILE}"')
        else:
            self.execute("; rm {TMP_FILE}")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_prepare(mod, env.TMP_DIR)
    assert os.path.exists(TMP_FILE)


def test_safe_open_src_unpack():
    """Tests that you can't use open in src_unpack"""
    file = """
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        with open("{BLOCKED_FILE}", "r") as file:
            file.read()
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_unpack(mod, env.TMP_DIR)


@pytest.mark.xfail(
    sys.platform == "win32",
    reason="Sandboxie's permissions don't allow nesting, so we have to use the "
    "default redirected writes, which don't raise exceptions",
)
def test_safe_open_src_prepare():
    """Tests that you can't write outside tmpdir with open in src_prepare"""
    file = """
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        with open("{TMP_FILE}", "w") as file:
            print("foo", file=file)
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    with pytest.raises(SandboxedError):
        src_prepare(mod, env.TMP_DIR)


def test_safe_open_allowed():
    """Tests situations where you can use open"""
    file = """
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        with open("{BLOCKED_FILE}", "r") as file:
            file.read()
        with open("{TMP_DIR}/foofile", "w") as file:
            print("foo", file=file)

    def src_unpack(self):
        with open("{TMP_DIR}/foofile", "w") as file:
            print("foo", file=file)
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_unpack(mod, env.TMP_DIR)
    src_prepare(mod, env.TMP_DIR)
    os.remove(f"{env.TMP_DIR}/foofile")


def test_stdout():
    """Tests that we can receive stdout from binary executables"""
    file = """
import os
import sys
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_prepare(self):
        output = self.execute("python --version", pipe_output=True)
        if not output.startswith("Python "):
            raise Exception("Incorrect output " + output)
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)


def test_git():
    """Tests that we can execute the git executable within the sandbox"""
    file = """
import os
import sys
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        self.execute("git --version")
        self.execute("git clone https://gitlab.com/portmod/blank.git")
        assert os.path.isdir("blank")
        assert os.path.exists(os.path.join("blank", "profiles", "repo_name"))

    def src_prepare(self):
        self.execute("git --version")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)
    src_unpack(mod, env.TMP_DIR)


def test_perl():
    """Tests that we can execute the perl executable within the sandbox"""
    if not shutil.which("perl"):
        pytest.skip("Perl not installed")

    file = """
import os
import sys
from pybuild import Pybuild1

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"

    def src_unpack(self):
        self.execute("perl --version")

    def src_prepare(self):
        self.execute("perl --version")
"""
    create_pybuild(file)
    mod = load_file(TMP_FILE)
    src_prepare(mod, env.TMP_DIR)
    src_unpack(mod, env.TMP_DIR)


def test_loader():
    _sandbox_execute_pybuild(
        os.path.join(TEST_REPO_DIR, "test/test/test-1.0.pybuild"),
        "cache",
        Permissions(),
    )


def test_inplacevar():
    """Tests that inplace modification of variables works"""
    file = """
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="test"
    LICENSE="GPL-3"

    def __init__(self):
        foo = 0
        foo += 10
        foo -= 1
        foo /= 2
        foo *= 2
        foo //= 2
        foo = 0
        foo |= 2
        foo ^= 2
        foo << 1
        foo >> 1
"""
    create_pybuild(file)
    __safe_load_file(TMP_FILE)


def test_augattribute():
    """Tests that augmented asignment of attributes works"""
    file = """
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="test"
    LICENSE="GPL-3"

    def __init__(self):
        self.foo = 0
        self.foo += 10
"""
    create_pybuild(file)
    __safe_load_file(TMP_FILE)


@pytest.mark.xfail(
    sys.platform == "win32",
    reason="unknown permissions error when cleaning up",
)
def test_path():
    """Tests that path modification from pybuilds with the exec property works"""
    file = """
import os
import stat
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="test"
    LICENSE="GPL-3"

    def src_install(self):
        os.makedirs(os.path.join(self.D, "bin"))
        path = os.path.join(self.D, "bin", "foo.py")
        with open(path, "w") as file:
            print("#!/usr/bin/env python", file=file)
            print("print(1)", file=file)
        os.chmod(path,
              stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
            | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
        )
"""
    create_pybuild(file)

    file2 = """
import shutil
from pybuild import Pybuild1


class Package(Pybuild1):
    NAME="Test"
    DESC="test"
    LICENSE="GPL-3"

    def src_prepare(self):
        # Note: we can't actually execute this as the test data may
        # be installed in a noexec filesystem, as /tmp and /var/tmp
        # sometimes are.
        shutil.which("foo.py")
"""

    pkg = load_file(TMP_FILE)
    install_pkg(pkg, set())

    create_pybuild(file2)
    pkg = load_file(TMP_FILE)
    src_prepare(pkg, env.TMP_DIR)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import logging

from .colour import bright, red, yellow
from .l10n import l10n


class PortmodFormatter(logging.Formatter):
    FORMATS = {
        logging.WARNING: logging.StrFormatStyle(
            bright(yellow("WARNING")) + ": {message}"
        ),
        logging.ERROR: logging.StrFormatStyle(bright(red("ERROR")) + ": {message}"),
        logging.CRITICAL: logging.StrFormatStyle(
            bright(red("CRITICAL")) + ": {message}"
        ),
        "DEFAULT": logging.StrFormatStyle("{message}"),
    }

    def __init__(self):
        super().__init__(style="{")

    def format(self, record):
        self._style = self.FORMATS.get(record.levelno, self.FORMATS["DEFAULT"])
        return logging.Formatter.format(self, record)


_LOG_HANDLER = None


def init_logger(args):
    """Initializes python logger"""
    global _LOG_HANDLER
    if not _LOG_HANDLER:
        _LOG_HANDLER = logging.StreamHandler()
        logging.root.addHandler(_LOG_HANDLER)
        formatter = PortmodFormatter()
        _LOG_HANDLER.setFormatter(formatter)

    if args.verbose:
        _LOG_HANDLER.setLevel(logging.DEBUG)
        logging.root.setLevel(logging.DEBUG)
    elif args.quiet:
        _LOG_HANDLER.setLevel(logging.WARN)
        logging.root.setLevel(logging.WARN)
    else:
        _LOG_HANDLER.setLevel(logging.INFO)
        logging.root.setLevel(logging.INFO)


def add_logging_arguments(parser):
    parser.add_argument("-q", "--quiet", help=l10n("quiet-help"), action="store_true")
    parser.add_argument(
        "-v", "--verbose", help=l10n("verbose-help"), action="store_true"
    )


__MESSAGES = set()


def warn_once(message: str):
    if message not in __MESSAGES:
        logging.warning(message)
        __MESSAGES.add(message)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for directly loading pybuild files.
These functions should not be called directly.
See portmod.loader for functions to load pybuilds safely using a sandbox.
"""

import ast
import glob
import importlib
import importlib.util
import os
import sys
from functools import lru_cache
from logging import warning
from types import SimpleNamespace
from typing import Any, Dict, Optional, cast

from RestrictedPython import (
    RestrictingNodeTransformer,
    compile_restricted_exec,
    limited_builtins,
    safe_globals,
)
from RestrictedPython.Eval import default_guarded_getitem, default_guarded_getiter
from RestrictedPython.Guards import (
    guarded_iter_unpack_sequence,
    guarded_unpack_sequence,
    safer_getattr,
)

from .atom import Atom
from .globals import vdb_path
from .pybuild import FullInstalledPybuild, FullPybuild

WHITELISTED_IMPORTS = {
    "filecmp",
    "glob",
    "os",
    "sys",
    "shutil",
    "stat",
    "os.path",
    "chardet",
    "pybuild",
    "pybuild.info",
    "pybuild.winreg",
    "re",
    "csv",
    "json",
    "typing",
    "fnmatch",
    "collections",
}


def find_installed_path(atom: Atom) -> Optional[str]:
    if atom.C:
        path = os.path.join(vdb_path(), atom.C, atom.PN)
        if os.path.exists(path):
            results = glob.glob(os.path.join(path, "*.pybuild"))
            assert len(results) == 1
            return results[0]
    else:
        for dirname in glob.glob(os.path.join(vdb_path(), "*")):
            path = os.path.join(vdb_path(), dirname, atom.PN)
            if os.path.exists(path):
                results = glob.glob(os.path.join(path, "*.pybuild"))
                assert len(results) == 1
                return results[0]
    return None


@lru_cache()
def _import_common(name: str) -> SimpleNamespace:
    """
    args:
        name: The import name as an absolute import path
        installed: Whether or not the package calling this is an installed package
        load_function: The function taking a file path and a keyword argument installed,
            indicating the installed status of the file to be loaded, to be used to load
            the common module

    returns:
        The Module as a SimpleNamespace
    """
    if len(name.split(".")) > 2:
        raise Exception(f"Invalid package {name}")
    _, module_name = name.split(".")
    base_atom = Atom(f"common/{module_name}")
    path = find_installed_path(base_atom)

    if not path:
        raise Exception(f"Could not find package {name}")

    result = __load_module(path, installed=True)

    return SimpleNamespace(
        **{key: value for key, value in result.items() if not key.startswith("_")}
    )


def _import(installed: bool, repo: Optional[str] = None):
    def import_fn(name, globs, loc, fromlist, level):
        if name.startswith("common."):
            return _import_common(name)
        if name not in WHITELISTED_IMPORTS:
            raise RuntimeError(f"Importing from {name} is not allowed")
        return __import__(name, globs, loc, fromlist, level)

    return import_fn


# Default implementation to handle invalid pybuilds
class Package:
    def __init__(self):
        raise Exception("Package is not defined")


def default_write_guard(ob):
    """Noop write guard"""
    return ob


def safer_hasattr(obj, name):
    """
    Version of hasattr implemented using safet_getattr

    This doesn't really provide any extra security, but does mean that
    str.format, and attributes beginning with underscores, which are
    blocked by safer_getattr, return False rather than True
    """
    try:
        safer_getattr(obj, name)
    except (NotImplementedError, AttributeError):
        return False
    return True


def default_apply(func, *args, **kwargs):
    return func(*args, **kwargs)


MINIMAL_GLOBALS: Dict[str, Any] = safe_globals
SAFE_GLOBALS: Dict[str, Any] = MINIMAL_GLOBALS.copy()
SAFE_GLOBALS["__builtins__"] = MINIMAL_GLOBALS["__builtins__"].copy()
SAFE_GLOBALS.update({"Package": Package})
SAFE_GLOBALS["__builtins__"].update(
    {
        "FileNotFoundError": FileNotFoundError,
        "__metaclass__": type,
        "_apply_": default_apply,
        "_getattr_": safer_getattr,
        "_getitem_": default_guarded_getitem,
        "_getiter_": default_guarded_getiter,
        "_iter_unpack_sequence_": guarded_iter_unpack_sequence,
        "_unpack_sequence_": guarded_unpack_sequence,
        "_write_": default_write_guard,
        "all": all,
        "any": any,
        "dict": dict,
        "enumerate": enumerate,
        "filter": filter,
        "frozenset": frozenset,
        "getattr": safer_getattr,
        "hasattr": safer_hasattr,
        "iter": iter,
        "map": map,
        "max": max,
        "min": min,
        "next": next,
        "reversed": reversed,
        "set": set,
        "sorted": sorted,
        "sum": sum,
        "super": super,
    }
)
SAFE_GLOBALS["__builtins__"].update(limited_builtins)

SANDBOX_GLOBALS: Dict[str, Any] = SAFE_GLOBALS.copy()
SANDBOX_GLOBALS["__builtins__"] = SAFE_GLOBALS["__builtins__"].copy()


class PrintWrapper:
    def __init__(self, _getattr_=None):
        self.txt = []
        self._getattr_ = _getattr_

    def write(self, text):
        self.txt.append(text)

    def __call__(self):
        return "".join(self.txt)

    def _call_print(self, *objects, **kwargs):
        if kwargs.get("file", None) is None:
            kwargs["file"] = sys.stdout
        else:
            self._getattr_(kwargs["file"], "write")
        print(*objects, **kwargs)


# print and open are only allowed within the sandbox
SANDBOX_GLOBALS["__builtins__"].update({"_print_": PrintWrapper, "open": open})


class Policy(RestrictingNodeTransformer):
    def visit_JoinedStr(self, node):
        return self.node_contents_visit(node)

    def visit_FormattedValue(self, node):
        return self.node_contents_visit(node)

    def visit_AnnAssign(self, node):
        return self.node_contents_visit(node)

    def visit_AugAssign(self, node):
        return self.node_contents_visit(node)

    def visit_ImportFrom(self, node):
        if (
            node.module in WHITELISTED_IMPORTS
            or node.module.startswith("pybuild")
            and node.level == 0
        ):
            try:
                module = importlib.import_module(node.module)
                for name in node.names:
                    if isinstance(module.__dict__.get(name.name), type(importlib)):
                        self.error(
                            node, "Importing modules from other modules is forbidden"
                        )
            except ModuleNotFoundError:
                pass

        return RestrictingNodeTransformer.visit_ImportFrom(self, node)

    def visit_FunctionDef(self, node):
        node = RestrictingNodeTransformer.visit_FunctionDef(self, node)
        if node.name == "__init__":
            newnode = ast.parse("super().__init__()").body[0]
            newnode.lineno = node.lineno
            newnode.col_offset = node.col_offset
            node.body.insert(0, newnode)
        return node


def restricted_load(code, filepath: str, _globals: Dict[str, Any]):
    if sys.platform == "win32":
        code = code.replace("\\", "\\\\")
    byte_code, errors, warnings, names = compile_restricted_exec(
        code, filename=filepath, policy=Policy
    )
    if errors:
        raise SyntaxError(errors)
    seen: Dict[str, str] = {}
    for message in [seen.setdefault(x, x) for x in warnings if x not in seen]:
        if not message.endswith("Prints, but never reads 'printed' variable."):
            warning(f"In file {filepath}: {message}")
    exec(byte_code, _globals, _globals)


def __load_module(path: str, *, installed=False) -> Dict[str, Any]:
    filename, _ = os.path.splitext(os.path.basename(path))

    with open(path, "r", encoding="utf-8") as file:
        code = file.read()
        tmp_globals = SANDBOX_GLOBALS.copy()
        tmp_globals["__builtins__"]["__import__"] = _import(
            installed=installed,
        )
        tmp_globals["__name__"] = filename
        restricted_load(code, path, tmp_globals)

    return tmp_globals


def __load_file_common(path: str, module: Dict[str, Any], installed: bool):
    module["Package"].__pybuild__ = path
    pkg = module["Package"]()
    pkg.FILE = os.path.abspath(path)
    pkg.INSTALLED = False

    if not installed:
        # determine common dependencies
        def find_common_imports(file: str):
            depends = []
            with open(file, "r", encoding="utf-8") as fp:
                tree = ast.parse(fp.read())

            def find_imports(tree: ast.AST):
                if isinstance(tree, ast.Module):
                    for statement in tree.body:
                        if isinstance(statement, ast.Import):
                            for alias in statement.names:
                                if alias.name.startswith("common."):
                                    depends.append(alias.name.replace(".", "/"))
                        elif isinstance(statement, ast.ImportFrom):
                            if statement.module and statement.module.startswith(
                                "common."
                            ):
                                depends.append(statement.module.replace(".", "/"))

            # Globals
            find_imports(tree)
            # TODO: Inline imports in functions?
            return depends

        pkg.RDEPEND = " ".join([pkg.RDEPEND] + find_common_imports(pkg.FILE))
    return pkg


def load_file(path: str, *, installed=False) -> FullPybuild:
    """
    Loads a pybuild file

    :param path: Path of the pybuild file
    """
    module = __load_module(path, installed=installed)
    pkg = __load_file_common(path, module, installed)
    return cast(FullPybuild, pkg)


def load_installed(file: str) -> FullInstalledPybuild:
    """
    Loads an installed pybuild

    :param file: Path of the pybuild file
    """
    mod = cast(FullInstalledPybuild, load_file(file, installed=True))
    __load_installed_common(mod, file)
    return mod


def __load_installed_common(mod, file: str):
    """Shared code for loading installed pybuilds"""
    mod.INSTALLED = True
    parent = os.path.dirname(file)

    def read_file(name: str) -> Optional[str]:
        if os.path.exists(os.path.join(parent, name)):
            with open(os.path.join(parent, name), "r") as repo_file:
                return repo_file.read().strip()
        return None

    repo = read_file("REPO")
    if not repo:
        raise Exception(
            f"Internal Error: Installed package in file {file}"
            "has no repository identifier"
        )
    mod.REPO = repo
    mod.INSTALLED_USE = set((read_file("USE") or "").split())
    mod.RDEPEND = read_file("RDEPEND") or mod.RDEPEND



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from functools import lru_cache
from types import SimpleNamespace
from typing import Dict, Optional, Union

from portmodlib.l10n import l10n

from ..globals import env


class BaseRepo(SimpleNamespace):
    def __init__(
        self,
        name: str,
        sync_type: Optional[str],
        sync_uri: Optional[str],
        priority: int,
    ):
        self.name = name
        self.sync_type = sync_type
        self.sync_uri = sync_uri
        self.priority = priority


class Repo(BaseRepo):
    def __init__(
        self,
        # Default is empty string so that pickling works.
        # Normally the name and location is required
        name: str = "",
        location: str = "",
        auto_sync: bool = False,
        sync_type: Optional[str] = None,
        sync_uri: Optional[str] = None,
        priority: int = 0,
    ):
        super().__init__(name, sync_type, sync_uri, priority)
        self.location = location
        self.auto_sync = auto_sync

    def to_dict(self) -> Dict[str, Union[str, int, bool]]:
        result = {}
        for attr in [
            "location",
            "auto_sync",
            "sync_type",
            "sync_uri",
            "priority",
        ]:
            if getattr(self, attr):
                result[attr] = getattr(self, attr)
        return result


class RemoteRepo(BaseRepo):
    def __init__(
        self,
        # Default is empty string so that pickling works.
        # Normally the name is required
        name: str = "",
        sync_type: Optional[str] = None,
        sync_uri: Optional[str] = None,
        priority: int = 0,
        description: str = "",
        quality: str = "unspecified",
    ):
        super().__init__(name, sync_type, sync_uri, priority)
        self.description = description
        self.quality = quality


@lru_cache()
def get_repo_name(path: str) -> str:
    """
    Given a path within a repo, returns the repo's name

    If path is not within a repo, returns None
    """
    root = get_repo_root(path)
    if root is not None:
        path = os.path.join(root, "profiles", "repo_name")
        if os.path.exists(path):
            with open(path, mode="r") as name_file:
                return name_file.read().strip()

    raise FileNotFoundError(
        f"Repo at path {path} does not contain the profiles/repo_name file"
    )


@lru_cache()
def get_repo_root(path: str) -> Optional[str]:
    """
    Returns the root given a path within a repository

    If the path is not in a repository, returns None
    """
    path = os.path.abspath(path)
    # Recursively look for metadata/repo_name to identify root
    if os.path.exists(os.path.join(path, "profiles", "repo_name")):
        return path
    if os.path.dirname(path) == path:
        # We've reached the root of the FS there is no repo
        return None

    return get_repo_root(os.path.dirname(path))


def has_repo(name: str) -> bool:
    """Returns true iff the repo exists"""
    for repo in env.prefix().REPOS:
        if repo.name == name:
            return True
    return False


def get_repo(name: str) -> Repo:
    """Returns repo of the given name"""
    if env.PREFIX_NAME:
        repos = env.prefix().REPOS
    else:
        repos = env.REPOS
    for repo in repos:
        if repo.name == name:
            return repo
    raise Exception(l10n("repo-does-not-exist", name=name))



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from functools import cmp_to_key
from types import SimpleNamespace
from typing import AbstractSet, Iterable, List, Mapping

from portmod.config.sets import is_selected
from portmod.config.use import get_use, get_user_global_use, get_user_use
from portmod.loader import load_all_installed, load_installed_pkg, load_pkg, load_pkg_fq
from portmod.repo import get_repo
from portmod.repo.metadata import get_use_flag_atom_aliases
from portmod.util import get_keyword_dep
from portmodlib.atom import Atom, FQAtom, QualifiedAtom, version_gt

from .formula import Formula, toggle, toggle_clause

WEIGHTS = SimpleNamespace()

###################################################################################
WEIGHTS.base_update = 1  # Base weight for the lowest
WEIGHTS.update_diff = 10  # Difference between each update, in order
# Weight to keep mods that are installed, but not selected
# The negative of this is set if depclean is enabled (plus total_up_weight)
WEIGHTS.keep_installed = 2
# Weight to keep a mod that is not selected and not installed disabled
WEIGHTS.not_needed = 2
WEIGHTS.user_flag = 4  # Weight to keep a user-set flag the same
WEIGHTS.default_flag = 3  # Weight to keep a flag at its default value
###################################################################################


def weigh_clauses(
    atoms: Mapping[QualifiedAtom, Iterable[FQAtom]],
    flags: Iterable[FQAtom],
    explicit: AbstractSet[Atom],
    *,
    deep: bool,
    depclean: bool,
    update: bool,
    newuse: bool,
    emptytree: bool,
) -> Formula:
    """Creates soft clauses for the given atoms"""
    formula = Formula()
    # Soft clauses to minimize unwanted changes to user's configuration

    # Track total weights so that we can make sure that higher-priority clauses
    # always have a greater weight
    total_up_weight = 0
    total_not_needed = 0

    # Clause for each mod not installed, with a small weight penalty for installing them
    for group in atoms:
        for atom in atoms[group]:
            if (
                not load_installed_pkg(Atom(atom.CPN))
                and not is_selected(atom)
                and Atom(atom.CPN) not in explicit
            ):
                formula.append(["-" + atom], weight=WEIGHTS.not_needed)
                total_not_needed += WEIGHTS.not_needed

    # When deep depcleaning, keep flags as they are,
    # so the only change should be packages being removed.
    flag_formula, total_flag_weight = weigh_flags(
        flags, atoms, explicit, newuse=newuse, freeze=depclean and deep and not newuse
    )
    formula.merge(flag_formula)

    # Ensure that the update weight is at least greater than the total number of packages to
    # overcome the basic resistance against moving from an installed package to the repository's
    # version of that installed package (i.e. rebuilding)
    total_atoms = sum(len(group) for group in atoms)

    # Penalize out of date versions.
    for group in atoms:
        # Divide mods into stable, testing and unstable
        # If user accepts testing keywords, they will be considered stable
        stable = set()
        testing = set()
        unstable = set()
        for atom in atoms[group]:
            mod = load_pkg_fq(atom)
            keyword = get_keyword_dep(mod)
            if not keyword:
                stable.add(atom)
            elif keyword.keyword.startswith("~"):
                testing.add(atom)
            elif keyword.keyword == "**":
                unstable.add(atom)

        i = WEIGHTS.base_update
        inc = WEIGHTS.update_diff + total_flag_weight

        def cmp(x, y):
            (vx, rx, ax) = x
            (vy, ry, ay) = y
            # Atoms are sorted first by version, then repo priority.
            # The installed Repo (here represented by None) has the
            # lowest priority
            if version_gt(vx, vy):
                return -1
            elif not version_gt(vy, vx):
                # If nether is greater than the other, they are equal
                # The installed version comes first, unless the package was selected
                if atom.CPN in explicit or emptytree:
                    if rx is None:
                        return 1
                    if ry is None:
                        return -1
                else:
                    if rx is None:
                        return -1
                    if ry is None:
                        return 1
                if rx.priority > ry.priority:
                    return -1
            return 1

        def weigh(li):
            nonlocal i, total_up_weight
            tosort = []
            for atom in li:
                if atom.R.endswith("::installed") or atom.R == "installed":
                    repo = None
                else:
                    repo = get_repo(atom.R)
                tosort.append((atom.PVR, repo, atom))

            inorder = sorted(tosort, key=cmp_to_key(cmp))
            for index, (version, repo, atom) in enumerate(inorder):
                formula.append(["-" + atom], weight=i)
                if (
                    index < len(inorder) - 1
                    and inorder[index + 1][0] == version
                    and repo is None
                ):
                    # This version is installed
                    # The weight difference between installed and non-installed should be exactly one
                    i += 1
                else:
                    i += inc + total_not_needed + total_atoms
                    total_up_weight += i

        weigh(stable)
        weigh(testing)
        weigh(unstable)

        # At most one version of the mod must be installed
        formula.append(
            atoms[group],
            "inviolable-rule",
            "At most one version may be installed",
            atmost=1,
        )

    total_depclean_weight = total_up_weight

    # Clause for each mod installed, but not selected, penalizing their removal
    # If depclean is set, we instead weigh them to remove if possible
    for mod in load_all_installed():
        if (
            not is_selected(mod.ATOM)
            and Atom(mod.CPN) not in explicit
            and not emptytree
        ):
            if update:  # If update is passed, penalize removal, but updates override
                formula.append([mod.ATOM], weight=WEIGHTS.keep_installed)
            elif depclean and deep:
                for pkg in load_pkg(Atom(mod.CPN)):
                    formula.append(
                        ["-" + pkg.ATOM],
                        weight=WEIGHTS.keep_installed + total_up_weight,
                    )
                    total_depclean_weight += total_up_weight
            else:  # if update is not passed, stay installed regardless of updates
                formula.append(
                    [mod.ATOM], weight=WEIGHTS.keep_installed + total_up_weight
                )
    # Packages which are already installed should, if not explicitly depcleaned,
    # be kep in their current state. This must override the weight of packages to be removed,
    # otherwise a package may be downgraded to avoid new dependencies.
    for pkg in load_all_installed():
        if (
            is_selected(pkg.ATOM)
            and Atom(pkg.CPN) not in explicit
            and not update
            and not emptytree
        ):
            formula.append(
                [pkg.ATOM], weight=WEIGHTS.keep_installed + total_depclean_weight
            )

    # When depcleaning, make sure the weight of installing not installed mods is higher
    # than the weight of removing unneeded mods. This prevents depcleaning from swapping out
    # mods in or expressions for each other.
    if depclean:
        for group in atoms:
            for atom in atoms[group]:
                if (
                    not load_installed_pkg(Atom(atom.CPN))
                    and not is_selected(atom)
                    and Atom(atom.CPN) not in explicit
                ):
                    formula.append(
                        ["-" + atom],
                        weight=WEIGHTS.not_needed
                        + total_up_weight
                        + WEIGHTS.keep_installed,
                    )
    return formula


def weigh_flags(
    flags: Iterable[FQAtom],
    atoms: Mapping[QualifiedAtom, Iterable[FQAtom]],
    explicit: AbstractSet[Atom],
    *,
    newuse: bool,
    freeze: bool,
):
    formula = Formula()

    def iterate_flags():
        for flagatom in flags:
            flag = list(flagatom.USE)[0]  # Note: each only has one flag
            enabledatom = flagatom.lstrip("-")
            atom_nouse = flagatom.strip_use()
            pkg = load_pkg_fq(atom_nouse)
            yield flag, atom_nouse, enabledatom, pkg

        # We also need to consider flags which don't show up in DEPEND-style strings
        # since flags in IUSE may have an alias
        for group in atoms:
            for atom in atoms[group]:
                pkg = load_pkg_fq(atom)
                for flag in pkg.IUSE_EFFECTIVE:
                    flagatom = FQAtom(f"{pkg.ATOM}[{flag}]")
                    if flagatom not in flags:
                        yield flag, pkg.ATOM, flagatom, pkg

    default_weight = WEIGHTS.default_flag
    total_flag_weight = 0
    total_user_weight = 0

    # Clauses for default use flag values, with a small penalty for changes from
    #    the default setting,
    repo_aliases = {}
    for flag, atom, flagatom, pkg in iterate_flags():
        if freeze:
            # All flags are required to keep their current state
            # Note: mod.get_use() will return installed use, not configured use,
            # for an installed package.
            enabled = pkg.get_use()

            if flag in enabled:
                formula.append(["-" + atom, flagatom])
            else:
                formula.append(["-" + atom, "-" + flagatom])

        elif newuse or atom.CPN in explicit or not load_installed_pkg(pkg.CPN):
            if pkg.REPO not in repo_aliases:
                repo_aliases[pkg.REPO] = get_use_flag_atom_aliases(
                    get_repo(pkg.REPO).location
                )
            aliases = repo_aliases[pkg.REPO]
            user_flags = get_user_use(pkg.ATOM)
            global_user_flags = get_user_global_use()
            if flag in user_flags:
                pass
            elif "-" + flag in user_flags:
                pass
            elif flag in global_user_flags:
                pass
            elif "-" + flag in global_user_flags:
                pass
            elif flag in aliases and aliases[flag].CPN != atom.CPN:
                alias = aliases[flag]

                if alias.CPN not in atoms:
                    # Package isn't installed, or even in the dependency tree,
                    # so default to disabled
                    formula.append(["-" + flagatom], weight=default_weight)
                    continue

                desc = [
                    "variable for linking alias to packages",
                    "Should never cause a conflict",
                ]

                aliasdeps: List[List[str]] = []
                for aliased_pkg in load_pkg(alias):
                    tokens: List[str] = []
                    if alias.USE:
                        # List of tokens relevant to the use flag, plus a variable which will be
                        # used later to reference it
                        # Var at beginning of clause represents the entire flag.
                        # I.e. if any of the parts are not present, it's false.
                        # If all the parts are present, it's true
                        tokens.append(formula.genvariable(desc))
                        for alias_flag in alias.USE:
                            if alias_flag.startswith("-"):
                                alias_flag.strip("-")
                                tokens.append(
                                    "-" + aliased_pkg.ATOM + "[" + alias_flag + "]"
                                )
                            else:
                                tokens.append(aliased_pkg.ATOM + "[" + alias_flag + "]")

                    tokens.append(aliased_pkg.ATOM)
                    aliasdeps.append(tokens)

                # If no package matching the alias is installed,
                # this will give weight to the flag being disabled.
                formula.append(
                    ["-" + flagatom] + [clause[0] for clause in aliasdeps],
                    weight=default_weight,
                )
                for clause in aliasdeps:
                    if len(clause) > 1:
                        # This makes the variable at the beginning true if all the parts are true
                        formula.append([clause[0]] + toggle_clause(clause[1:]))

                        # This makes the variable at the beginning false if any of the parts are false
                        for term in clause[1:]:
                            formula.append(["-" + clause[0], term])

                # var is true iff no packages matching the alias are installed
                var = formula.genvariable(desc)
                # This will give weight to the flag being enabled, unless no package
                # matching the alias is installed.
                formula.append([flagatom, var], weight=default_weight)

                # If all packages are not installed, var must be true
                formula.append(
                    [var] + [clause[0] for clause in aliasdeps],
                )

                for clause in aliasdeps:
                    # If package is installed, var must be false
                    formula.append(["-" + var, toggle(clause[0])])

            elif flag in get_use(pkg)[0]:  # default value
                formula.append(["-" + atom, flagatom], weight=default_weight)
                total_flag_weight += default_weight
            else:
                formula.append(["-" + atom, "-" + flagatom], weight=default_weight)
                total_flag_weight += default_weight
        else:
            # If not running in deep mode, keep flags the same for
            # non-selected but previously installed packages, if possible
            enabled = pkg.get_use()

            if flag in enabled:
                formula.append(["-" + atom, flagatom], weight=default_weight)
            else:
                formula.append(["-" + atom, "-" + flagatom], weight=default_weight)

    user_weight = WEIGHTS.user_flag + total_flag_weight
    # Clauses for user flags. These will always override changes from the default value.
    for flag, atom, flagatom, mod in iterate_flags():
        if newuse or atom.CPN in explicit or not load_installed_pkg(mod.CPN):
            user_flags = get_user_use(mod.ATOM)
            global_user_flags = get_user_global_use()
            if flag in user_flags:
                formula.append(["-" + atom, flagatom], weight=user_weight)
                total_user_weight += user_weight
            elif "-" + flag in user_flags:
                formula.append(["-" + atom, "-" + flagatom], weight=user_weight)
                total_user_weight += user_weight
            elif flag in global_user_flags:
                formula.append(["-" + atom, flagatom], weight=user_weight)
                total_user_weight += user_weight
            elif "-" + flag in global_user_flags:
                formula.append(["-" + atom, "-" + flagatom], weight=user_weight)
                total_user_weight += user_weight

    return formula, total_user_weight



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from typing import Dict

from portmodlib.fs import get_hash
from portmodlib.l10n import l10n
from portmodlib.source import HashAlg, Source


class LocalHashError(Exception):
    """Exception indicating an unexpected download file"""


class SourceManifest(Source):
    """Class used for storing information about download files"""

    def __init__(self, source: Source, hashes: Dict[HashAlg, str], size: int):
        super().__init__(source.url, source.name)
        self.hashes = hashes
        self.size = size

    def __hash__(self):
        return hash((self.url, self.name, tuple(self.hashes)))

    def as_source(self):
        return Source(self.url, self.name)

    def check_file(self, filename: str, raise_ex=False) -> bool:
        """
        Returns true if and only if the hash of the given file
        matches the stored hash
        """

        hashes_to_check = sorted(self.hashes)
        if len(hashes_to_check) > 1 and HashAlg.MD5 in hashes_to_check:
            # Ignore MD5 unless it's the only hash. It's neither particularly fast, nor reliable
            # It's only used since certain services will supply an MD5 hash which we can compare to
            hashes_to_check.remove(HashAlg.MD5)
        results = get_hash(filename, tuple(sorted(self.hashes)))
        for halg, result in zip(sorted(self.hashes), results):
            if self.hashes[halg] != result:
                if raise_ex:
                    raise LocalHashError(
                        l10n(
                            "local-hash-mismatch",
                            filename=filename,
                            hash=halg.name,
                            hash1=self.hashes[halg],
                            hash2=result,
                        )
                    )
                return False
        return True
